# A copy of napari.layers.shapes._shapes
from copy import copy

import numpy as np

from ._bounding_box_constants import Box, Mode


def highlight(layer, event):
    """Highlight hovered bounding boxes."""
    layer._set_highlight()


def select(layer, event):
    """Select bounding boxes or vertices either in select or direct select mode.

    Once selected bounding boxes can be moved or resized, and vertices can be moved
    depending on the mode. Holding shift when resizing a bounding box will preserve
    the aspect ratio.
    """
    shift = 'Shift' in event.modifiers
    # on press
    value = layer.get_value(event.position, world=True)
    layer._moving_value = copy(value)
    bounding_box_under_cursor, vertex_under_cursor = value
    if vertex_under_cursor is None:
        if shift and bounding_box_under_cursor is not None:
            if bounding_box_under_cursor in layer.selected_data:
                layer.selected_data.remove(bounding_box_under_cursor)
            else:
                layer.selected_data.add(bounding_box_under_cursor)
        elif bounding_box_under_cursor is not None:
            if bounding_box_under_cursor not in layer.selected_data:
                layer.selected_data = {bounding_box_under_cursor}
        else:
            layer.selected_data = set()
    layer._set_highlight()

    # we don't update the thumbnail unless a bounding box has been moved
    update_thumbnail = False
    yield

    # on move
    while event.type == 'mouse_move':
        coordinates = layer.world_to_data(event.position)
        # ToDo: Need to pass moving_coordinates to allow fixed aspect ratio
        # keybinding to work, this should be dropped
        layer._moving_coordinates = coordinates
        # Drag any selected bounding boxes
        if len(layer.selected_data) == 0:
            _drag_selection_box(layer, coordinates)
        else:
            _move(layer, coordinates)

        # if a bounding box is being moved, update the thumbnail
        if layer._is_moving:
            update_thumbnail = True
        yield

    # on release
    shift = 'Shift' in event.modifiers
    if not layer._is_moving and not layer._is_selecting and not shift:
        if bounding_box_under_cursor is not None:
            layer.selected_data = {bounding_box_under_cursor}
        else:
            layer.selected_data = set()
    elif layer._is_selecting:
        layer.selected_data = layer._data_view.bounding_boxes_in_box(layer._drag_box)
        layer._is_selecting = False
        layer._set_highlight()

    layer._is_moving = False
    layer._drag_start = None
    layer._drag_box = None
    layer._fixed_vertex = None
    layer._moving_value = (None, None)
    layer._set_highlight()

    if update_thumbnail:
        layer._update_thumbnail()


def add_bounding_box(layer, event):
    """Add a rectangle."""
    size = layer._vertex_size * layer.scale_factor / 4
    coordinates = layer.world_to_data(event.position)
    corner = np.array(coordinates).reshape(1, -1)

    data = np.tile(corner, (2**layer.ndim, 1))
    sizes = np.eye(layer.ndim)*size
    s_idx = [[]]
    for d in range(layer.ndim):
        s_idx.extend(list(e.copy()+[d] for e in s_idx))
    addition = np.asarray([sizes[idx].sum(0) for idx in s_idx])
    data = data + addition
    data = data[np.newaxis]
    yield from _add_bounding_box(
        layer, event, data=data
    )


def _add_bounding_box(layer, event, data):
    """Helper function for adding a a line, rectangle or ellipse."""
    # on press
    # Start drawing rectangle / ellipse / line
    layer.add(data)
    layer.selected_data = {layer.nbounding_boxes - 1}
    layer._value = (layer.nbounding_boxes - 1, 4)
    layer._moving_value = copy(layer._value)
    layer.refresh()
    yield

    data = layer.data[layer.nbounding_boxes - 1]
    # on move
    while event.type == 'mouse_move':
        # Drag any selected bounding boxes
        coordinates = layer.world_to_data(event.position)
        _move(layer, coordinates)
        min = data.min(0)
        max = data.max(0)
        size = max-min
        visible_size = size[layer._dims_displayed]
        max[layer._dims_displayed] = np.nan
        min[layer._dims_displayed] = np.nan
        # print("original data", data)
        data[:] = np.where(data == max, coordinates+visible_size.mean()/2, data)
        data[:] = np.where(data == min, coordinates-visible_size.mean()/2, data)
        # print("original size", size)
        # print("visible size", visible_size)
        # print("coordinates", coordinates)
        # print("new data", data)
        yield

    # on release
    layer._finish_drawing()


def _drag_selection_box(layer, coordinates):
    """Drag a selection box.

    Parameters
    ----------
    layer : .bounding_boxes.BoundingBoxLayer
        Bounding Box layer.
    coordinates : tuple
        Position of mouse cursor in data coordinates.
    """
    # If something selected return
    if len(layer.selected_data) > 0:
        return

    coord = [coordinates[i] for i in layer._dims_displayed]

    # Create or extend a selection box
    layer._is_selecting = True
    if layer._drag_start is None:
        layer._drag_start = coord
    layer._drag_box = np.array([layer._drag_start, coord])
    layer._set_highlight()


def _move(layer, coordinates):
    """Moves object at given mouse position and set of indices.

    Parameters
    ----------
    layer : BoundingBoxLayer
        BoundingBoxLayer layer.
    coordinates : tuple
        Position of mouse cursor in data coordinates.
    """
    # If nothing selected return
    if len(layer.selected_data) == 0:
        return

    vertex = layer._moving_value[1]

    if layer._mode in (
        [Mode.SELECT, Mode.ADD_BOUNDING_BOX]
    ):
        coord = [coordinates[i] for i in layer._dims_displayed]
        layer._moving_coordinates = coordinates
        layer._is_moving = True
        if vertex is None:
            # Check where dragging box from to move whole object
            if layer._drag_start is None:
                center = layer._selected_box[Box.CENTER]
                layer._drag_start = coord - center
            center = layer._selected_box[Box.CENTER]
            shift = coord - center - layer._drag_start
            for index in layer.selected_data:
                layer._data_view.shift(index, shift)
            layer._selected_box = layer._selected_box + shift
            layer.refresh()
        elif vertex < Box.LEN:
            # Corner / edge vertex is being dragged so resize object
            box = layer._selected_box
            if layer._fixed_vertex is None:
                layer._fixed_index = (vertex + 4) % Box.LEN
                layer._fixed_vertex = box[layer._fixed_index]

            size = (
                box[(layer._fixed_index + 4) % Box.LEN]
                - box[layer._fixed_index]
            )
            offset = box[Box.HANDLE] - box[Box.CENTER]
            if np.linalg.norm(offset) == 0:
                offset = [1, 1]
            offset = offset / np.linalg.norm(offset)
            offset_perp = np.array([offset[1], -offset[0]])

            fixed = layer._fixed_vertex
            new = list(coord)

            if layer._fixed_aspect and layer._fixed_index % 2 == 0:
                if (new - fixed)[0] == 0:
                    ratio = 1
                else:
                    ratio = abs((new - fixed)[1] / (new - fixed)[0])
                if ratio > layer._aspect_ratio:
                    r = layer._aspect_ratio / ratio
                    new[1] = fixed[1] + (new[1] - fixed[1]) * r
                else:
                    r = ratio / layer._aspect_ratio
                    new[0] = fixed[0] + (new[0] - fixed[0]) * r

            # if size @ offset == 0:
            if np.allclose(size @ offset, 0):
                dist = 1
            else:
                dist = ((new - fixed) @ offset) / (size @ offset)

            if size @ offset_perp == 0:
                dist_perp = 1
            else:
                dist_perp = ((new - fixed) @ offset_perp) / (
                    size @ offset_perp
                )

            if layer._fixed_index % 2 == 0:
                # corner selected
                scale = np.array([dist_perp, dist])
            elif layer._fixed_index % 4 == 3:
                # top selected
                scale = np.array([1, dist])
            else:
                # side selected
                scale = np.array([dist_perp, 1])

            # prevent box from shrinking below a threshold size
            threshold = layer._vertex_size * layer.scale_factor / 8
            scale[abs(scale * size[[1, 0]]) < threshold] = 1

            # check orientation of box
            angle = -np.arctan2(offset[0], -offset[1])
            c, s = np.cos(angle), np.sin(angle)
            # if angle == 0:
            if np.allclose(angle, 0):
                for index in layer.selected_data:
                    layer._data_view.scale(
                        index, scale, center=layer._fixed_vertex
                    )
                layer._scale_box(scale, center=layer._fixed_vertex)
            else:
                rotation = np.array([[c, s], [-s, c]])
                scale_mat = np.array([[scale[0], 0], [0, scale[1]]])
                inv_rot = np.array([[c, -s], [s, c]])
                transform = rotation @ scale_mat @ inv_rot
                for index in layer.selected_data:
                    layer._data_view.shift(index, -layer._fixed_vertex)
                    layer._data_view.transform(index, transform)
                    layer._data_view.shift(index, layer._fixed_vertex)
                layer._transform_box(transform, center=layer._fixed_vertex)
            layer.refresh()
        elif vertex == 8:
            # Rotation handle is being dragged so rotate object
            handle = layer._selected_box[Box.HANDLE]
            if layer._drag_start is None:
                layer._fixed_vertex = layer._selected_box[Box.CENTER]
                offset = handle - layer._fixed_vertex
                layer._drag_start = -np.degrees(
                    np.arctan2(offset[0], -offset[1])
                )

            new_offset = coord - layer._fixed_vertex
            new_angle = -np.degrees(np.arctan2(new_offset[0], -new_offset[1]))
            fixed_offset = handle - layer._fixed_vertex
            fixed_angle = -np.degrees(
                np.arctan2(fixed_offset[0], -fixed_offset[1])
            )

            if np.linalg.norm(new_offset) < 1:
                angle = 0
            elif layer._fixed_aspect:
                angle = np.round(new_angle / 45) * 45 - fixed_angle
            else:
                angle = new_angle - fixed_angle

            for index in layer.selected_data:
                layer._data_view.rotate(
                    index, angle, center=layer._fixed_vertex
                )
            # layer._rotate_box(angle, center=layer._fixed_vertex)
            layer.refresh()
    elif layer._mode == Mode.DIRECT:
        if vertex is not None:
            layer._is_moving = True
            index = layer._moving_value[0]
            vertices = layer._data_view.bounding_boxes[index].data
            vertices[vertex] = coordinates
            layer._data_view.edit(index, vertices)
            bounding_boxes = layer.selected_data
            layer._selected_box = layer.interaction_box(bounding_boxes)
            layer.refresh()



# A copy of napari.layers.shapes.shapes
import warnings
from contextlib import contextmanager
from copy import deepcopy, copy
from itertools import cycle
from typing import Dict, Optional, Union, Tuple
import napari.layers
import pandas as pd
from napari.layers import Layer
from napari.layers.utils.color_manager_utils import map_property, guess_continuous
from napari.layers.utils.color_transformations import transform_color_with_defaults, ColorType, \
    normalize_and_broadcast_colors, transform_color_cycle
from napari.layers.utils.layer_utils import dataframe_to_properties, _FeatureTable
from napari.layers.utils.text_manager import TextManager
from napari.utils import Colormap
from napari.utils.colormaps import ensure_colormap, ValidColormapArg
from napari.utils.colormaps.standardize_color import transform_color, rgb_to_hex, hex_to_name
from napari.utils.events import Event
from napari.utils.events.custom_types import Array
from napari.utils.misc import ensure_iterable
from vispy.color import get_color_names


from ._bounding_box_constants import ColorMode, BACKSPACE, Box
from ._bounding_box_list import BoundingBoxList
from .bounding_box import BoundingBox
from ._bounding_box_mouse_bindings import select, highlight, add_bounding_box
# from ._bounding_boxes_key_bindings import *
from .qt_bounding_box_control import *
from .vispy_bounding_box_layer import *
from ._bounding_box_utils import create_box

DEFAULT_COLOR_CYCLE = np.array([[1, 0, 1, 1], [0, 1, 0, 1]])

class BoundingBoxLayer(Layer):
    _colors = get_color_names()
    _vertex_size = 10
    _rotation_handle_length = 20
    _highlight_color = (0, 0.6, 1)
    _highlight_width = 1.5
    _max_bounding_boxes_thumbnail = 100
    def __init__(self,
                 data=None,
                 *,

                 ndim=None,
                 features=None,
                 properties=None,
                 property_choices=None,
                 text=None,
                 edge_width=1,
                 edge_color='black',
                 edge_color_cycle=None,
                 edge_colormap='viridis',
                 edge_contrast_limits=None,
                 face_color='white',
                 face_color_cycle=None,
                 face_colormap='viridis',
                 face_contrast_limits=None,
                 z_index=0,
                 name=None,
                 metadata=None,
                 scale=None,
                 translate=None,
                 rotate=None,
                 shear=None,
                 affine=None,
                 opacity=0.7,
                 blending='translucent',
                 visible=True,
    ):
        if data is None:
            if ndim is None:
                ndim = 2
            data = np.empty((0, 2, ndim))
        else:
            data = np.asarray(data)
            data_ndim = data.shape[-1]
            if ndim is not None and ndim != data_ndim:
                raise ValueError(
                    trans._(
                        "Bounding box dimensions must be equal to ndim",
                        deferred=True,
                    )
                )
            ndim = data_ndim
        super().__init__(
            data,
            ndim=ndim,
            name=name,
            metadata=metadata,
            scale=scale,
            translate=translate,
            rotate=rotate,
            shear=shear,
            affine=affine,
            opacity=opacity,
            blending=blending,
            visible=visible,
        )
        self.events.add(
            mode=Event,
            edge_width=Event,
            edge_color=Event,
            face_color=Event,
            properties=Event,
            current_edge_color=Event,
            current_face_color=Event,
            current_properties=Event,
            highlight=Event,
        )
        self._allow_thumbnail_update = True

        self._display_order_stored = []
        self._ndisplay_stored = self._ndisplay

        self._feature_table = _FeatureTable.from_layer(
            features=features,
            properties=properties,
            property_choices=property_choices,
            num_data=len(data),
        )
        # self.mouse_drag_callbacks.append(self._on_click)
        # Save the properties

        if np.isscalar(edge_width):
            self._current_edge_width = edge_width
        else:
            self._current_edge_width = 1

        self._data_view = BoundingBoxList(ndisplay=self._ndisplay)
        self._data_view.slice_key = np.array(self._slice_indices)[
            list(self._dims_not_displayed)
        ]

        self._value = (None, None)
        self._value_stored = (None, None)
        self._moving_value = (None, None)
        self._selected_data = set()
        self._selected_data_stored = set()
        self._selected_data_history = set()
        self._selected_box = None

        self._drag_start = None
        self._fixed_vertex = None
        self._fixed_aspect = False
        self._aspect_ratio = 1
        self._is_moving = False
        # _moving_coordinates are needed for fixing aspect ratio during
        # a resize
        self._moving_coordinates = None
        self._fixed_index = 0
        self._is_selecting = False
        self._drag_box = None
        self._drag_box_stored = None
        self._is_creating = False
        self._clipboard = {}

        self._mode = None
        self.mode = Mode.PAN_ZOOM
        self._status = self.mode

        self._init_bounding_boxes(
            data,
            edge_width=edge_width,
            edge_color=edge_color,
            edge_color_cycle=edge_color_cycle,
            edge_colormap=edge_colormap,
            edge_contrast_limits=edge_contrast_limits,
            face_color=face_color,
            face_color_cycle=face_color_cycle,
            face_colormap=face_colormap,
            face_contrast_limits=face_contrast_limits,
            z_index=z_index,
        )

        # set the current_* properties
        if len(data) > 0:
            self._current_edge_color = self.edge_color[-1]
            self._current_face_color = self.face_color[-1]
        elif len(data) == 0 and self.properties:
            self._initialize_current_color_for_empty_layer(edge_color, 'edge')
            self._initialize_current_color_for_empty_layer(face_color, 'face')
        elif len(data) == 0 and len(self.properties) == 0:
            self._current_edge_color = transform_color_with_defaults(
                num_entries=1,
                colors=edge_color,
                elem_name="edge_color",
                default="black",
            )
            self._current_face_color = transform_color_with_defaults(
                num_entries=1,
                colors=face_color,
                elem_name="face_color",
                default="black",
            )
            self.current_properties = {}

        self._text = TextManager._from_layer(
            text=text,
            n_text=self.nbounding_boxes,
            properties=self.properties,
        )

        # Trigger generation of view slice and thumbnail
        self._update_dims()
        self._mouse_down = False

    @property
    def features(self):
        """Dataframe-like features table.

        It is an implementation detail that this is a `pandas.DataFrame`. In the future,
        we will target the currently-in-development Data API dataframe protocol [1].
        This will enable us to use alternate libraries such as xarray or cuDF for
        additional features without breaking existing usage of this.

        If you need to specifically rely on the pandas API, please coerce this to a
        `pandas.DataFrame` using `features_to_pandas_dataframe`.

        References
        ----------
        .. [1]: https://data-apis.org/dataframe-protocol/latest/API.html
        """
        return self._feature_table.values

    @features.setter
    def features(
            self,
            features: Union[Dict[str, np.ndarray], pd.DataFrame],
    ) -> None:
        self._feature_table.set_values(features, num_data=self.nshapes)
        if self._face_color_property and (
                self._face_color_property not in self.features
        ):
            self._face_color_property = ''
            warnings.warn(
                trans._(
                    'property used for face_color dropped',
                    deferred=True,
                ),
                RuntimeWarning,
            )

        if self._edge_color_property and (
                self._edge_color_property not in self.features
        ):
            self._edge_color_property = ''
            warnings.warn(
                trans._(
                    'property used for edge_color dropped',
                    deferred=True,
                ),
                RuntimeWarning,
            )

        if self.text.values is not None:
            self.refresh_text()
        self.events.properties()

    @property
    def feature_defaults(self):
        """Dataframe-like with one row of feature default values.

        See `features` for more details on the type of this property.
        """
        return self._feature_table.defaults

    @property
    def properties(self) -> Dict[str, np.ndarray]:
        """dict {str: np.ndarray (N,)}, DataFrame: Annotations for each shape"""
        return self._feature_table.properties()

    @properties.setter
    def properties(self, properties: Dict[str, Array]):
        self.features = properties

    @property
    def property_choices(self) -> Dict[str, np.ndarray]:
        return self._feature_table.choices()

    def _initialize_current_color_for_empty_layer(
            self, color: ColorType, attribute: str
    ):
        """Initialize current_{edge,face}_color when starting with empty layer.

        Parameters
        ----------
        color : (N, 4) array or str
            The value for setting edge or face_color
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        """
        color_mode = getattr(self, f'_{attribute}_color_mode')
        if color_mode == ColorMode.DIRECT:
            curr_color = transform_color_with_defaults(
                num_entries=1,
                colors=color,
                elem_name=f'{attribute}_color',
                default="white",
            )

        elif color_mode == ColorMode.CYCLE:
            color_cycle = getattr(self, f'_{attribute}_color_cycle')
            curr_color = transform_color(next(color_cycle))

            # add the new color cycle mapping
            color_property = getattr(self, f'_{attribute}_color_property')
            prop_value = self._property_choices[color_property][0]
            color_cycle_map = getattr(self, f'{attribute}_color_cycle_map')
            color_cycle_map[prop_value] = np.squeeze(curr_color)
            setattr(self, f'{attribute}_color_cycle_map', color_cycle_map)

        elif color_mode == ColorMode.COLORMAP:
            color_property = getattr(self, f'_{attribute}_color_property')
            prop_value = self._property_choices[color_property][0]
            colormap = getattr(self, f'{attribute}_colormap')
            contrast_limits = getattr(self, f'_{attribute}_contrast_limits')
            curr_color, _ = map_property(
                prop=prop_value,
                colormap=colormap,
                contrast_limits=contrast_limits,
            )
        setattr(self, f'_current_{attribute}_color', curr_color)

    def _validate_properties(
        self, properties: Dict[str, np.ndarray], n_bounding_boxes: Optional[int] = None
    ) -> Dict[str, np.ndarray]:
        """Validates the type and size of the properties"""
        if n_bounding_boxes is None:
            n_bounding_boxes = len(self.data)
        for k, v in properties.items():
            if len(v) != n_bounding_boxes:
                raise ValueError(
                    trans._(
                        'the number of properties must equal the number of bounding boxes',
                        deferred=True,
                    )
                )
            # ensure the property values are a numpy array
            if type(v) != np.ndarray:
                properties[k] = np.asarray(v)

        return properties

    @property
    def data(self):
        """list: Each element is an (N, D) array of the vertices of a bounding box."""
        return self._data_view.data

    @data.setter
    def data(self, data):
        self._finish_drawing()

        n_new_bounding_boxes = len(data)

        edge_widths = self._data_view.edge_widths
        edge_color = self._data_view.edge_color
        face_color = self._data_view.face_color
        z_indices = self._data_view.z_indices

        # fewer bounding boxes, trim attributes
        if self.nbounding_boxes > n_new_bounding_boxes:
            edge_widths = edge_widths[:n_new_bounding_boxes]
            z_indices = z_indices[:n_new_bounding_boxes]
            edge_color = edge_color[:n_new_bounding_boxes]
            face_color = face_color[:n_new_bounding_boxes]
        # more bounding boxes, add attributes
        elif self.nbounding_boxes < n_new_bounding_boxes:
            n_bounding_boxes_difference = n_new_bounding_boxes - self.nbounding_boxes
            edge_widths = edge_widths + [1] * n_bounding_boxes_difference
            z_indices = z_indices + [0] * n_bounding_boxes_difference
            edge_color = np.concatenate(
                (
                    edge_color,
                    self._get_new_bounding_box_color(n_bounding_boxes_difference, 'edge'),
                )
            )
            face_color = np.concatenate(
                (
                    face_color,
                    self._get_new_bounding_box_color(n_bounding_boxes_difference, 'face'),
                )
            )

        self._data_view = BoundingBoxList()
        self._data_view.slice_key = np.array(self._slice_indices)[
            list(self._dims_not_displayed)
        ]
        self.add(
            data,
            edge_width=edge_widths,
            edge_color=edge_color,
            face_color=face_color,
            z_index=z_indices,
        )

        self._update_dims()
        self.events.data(value=self.data)
        self._set_editable()

    @property
    def _extent_data(self) -> np.ndarray:
        if len(self.data) == 0:
            extrema = np.full((2, self.ndim), np.nan)
        else:
            maxs = np.max([np.max(d, axis=0) for d in self.data], axis=0)
            mins = np.min([np.min(d, axis=0) for d in self.data], axis=0)
            extrema = np.vstack([mins, maxs])
        return extrema

    @property
    def current_edge_width(self):
        """float: Width of bounding box edges including lines and paths."""
        return self._current_edge_width

    @current_edge_width.setter
    def current_edge_width(self, edge_width):
        self._current_edge_width = edge_width
        if self._update_properties:
            for i in self.selected_data:
                self._data_view.update_edge_width(i, edge_width)
        self.events.edge_width()

    @property
    def selected_data(self):
        """set: set of currently selected bounding boxes."""
        return self._selected_data

    @selected_data.setter
    def selected_data(self, selected_data):
        self._selected_data = set(selected_data)
        self._selected_box = self.interaction_box(self._selected_data)

        # Update properties based on selected bounding boxes
        if len(selected_data) > 0:
            selected_data_indices = list(selected_data)
            selected_face_colors = self._data_view._face_color[
                selected_data_indices
            ]
            face_colors = np.unique(selected_face_colors, axis=0)
            if len(face_colors) == 1:
                face_color = face_colors[0]
                with self.block_update_properties():
                    self.current_face_color = face_color

            selected_edge_colors = self._data_view._edge_color[
                selected_data_indices
            ]
            edge_colors = np.unique(selected_edge_colors, axis=0)
            if len(edge_colors) == 1:
                edge_color = edge_colors[0]
                with self.block_update_properties():
                    self.current_edge_color = edge_color

            edge_width = list(
                {self._data_view.bounding_boxes[i].edge_width for i in selected_data}
            )
            if len(edge_width) == 1:
                edge_width = edge_width[0]
                with self.block_update_properties():
                    self.current_edge_width = edge_width
            properties = {}
            for k, v in self.properties.items():
                # pandas uses `object` as dtype for strings by default, which
                # combined with the axis argument breaks np.unique
                axis = 0 if v.ndim > 1 else None
                properties[k] = np.unique(v[selected_data_indices], axis=axis)
            n_unique_properties = np.array(
                [len(v) for v in properties.values()]
            )
            if np.all(n_unique_properties == 1):
                with self.block_update_properties():
                    self.current_properties = properties

    @property
    def text(self) -> TextManager:
        """TextManager: The TextManager object containing the text properties"""
        return self._text

    @text.setter
    def text(self, text):
        self._text._update_from_layer(
            text=text,
            n_text=self.nbounding_boxes,
            properties=self.properties,
        )

    @property
    def edge_color(self):
        """(N x 4) np.ndarray: Array of RGBA face colors for each bounding box"""
        return self._data_view.edge_color

    @edge_color.setter
    def edge_color(self, edge_color):
        self._set_color(edge_color, 'edge')
        self.events.edge_color()
        self._update_thumbnail()

    @property
    def edge_color_cycle(self) -> np.ndarray:
        """Union[list, np.ndarray] :  Color cycle for edge_color.

        Can be a list of colors defined by name, RGB or RGBA
        """
        return self._edge_color_cycle_values

    @edge_color_cycle.setter
    def edge_color_cycle(self, edge_color_cycle: Union[list, np.ndarray]):
        self._set_color_cycle(edge_color_cycle, 'edge')

    @property
    def edge_colormap(self) -> Tuple[str, Colormap]:
        """Return the colormap to be applied to a property to get the edge color.

        Returns
        -------
        colormap : napari.utils.Colormap
            The Colormap object.
        """
        return self._edge_colormap

    @edge_colormap.setter
    def edge_colormap(self, colormap: ValidColormapArg):
        self._edge_colormap = ensure_colormap(colormap)

    @property
    def edge_contrast_limits(self) -> Tuple[float, float]:
        """None, (float, float): contrast limits for mapping
        the edge_color colormap property to 0 and 1
        """
        return self._edge_contrast_limits

    @edge_contrast_limits.setter
    def edge_contrast_limits(
            self, contrast_limits: Union[None, Tuple[float, float]]
    ):
        self._edge_contrast_limits = contrast_limits

    @property
    def edge_color_mode(self) -> str:
        """str: Edge color setting mode

        DIRECT (default mode) allows each bounding box color to be set arbitrarily

        CYCLE allows the color to be set via a color cycle over an attribute

        COLORMAP allows color to be set via a color map over an attribute
        """
        return str(self._edge_color_mode)

    @edge_color_mode.setter
    def edge_color_mode(self, edge_color_mode: Union[str, ColorMode]):
        self._set_color_mode(edge_color_mode, 'edge')

    @property
    def face_color(self):
        """(N x 4) np.ndarray: Array of RGBA face colors for each bounding box"""
        return self._data_view.face_color

    @face_color.setter
    def face_color(self, face_color):
        self._set_color(face_color, 'face')
        self.events.face_color()
        self._update_thumbnail()

    @property
    def face_color_cycle(self) -> np.ndarray:
        """Union[np.ndarray, cycle]:  Color cycle for face_color
        Can be a list of colors defined by name, RGB or RGBA
        """
        return self._face_color_cycle_values

    @face_color_cycle.setter
    def face_color_cycle(self, face_color_cycle: Union[np.ndarray, cycle]):
        self._set_color_cycle(face_color_cycle, 'face')

    @property
    def face_colormap(self) -> Tuple[str, Colormap]:
        """Return the colormap to be applied to a property to get the face color.

        Returns
        -------
        colormap : napari.utils.Colormap
            The Colormap object.
        """
        return self._face_colormap

    @face_colormap.setter
    def face_colormap(self, colormap: ValidColormapArg):
        self._face_colormap = ensure_colormap(colormap)

    @property
    def face_contrast_limits(self) -> Union[None, Tuple[float, float]]:
        """None, (float, float) : clims for mapping the face_color
        colormap property to 0 and 1
        """
        return self._face_contrast_limits

    @face_contrast_limits.setter
    def face_contrast_limits(
            self, contrast_limits: Union[None, Tuple[float, float]]
    ):
        self._face_contrast_limits = contrast_limits

    @property
    def face_color_mode(self) -> str:
        """str: Face color setting mode

        DIRECT (default mode) allows each bounding box color to be set arbitrarily

        CYCLE allows the color to be set via a color cycle over an attribute

        COLORMAP allows color to be set via a color map over an attribute
        """
        return str(self._face_color_mode)

    @face_color_mode.setter
    def face_color_mode(self, face_color_mode):
        self._set_color_mode(face_color_mode, 'face')

    @property
    def current_edge_color(self):
        """str: color of bounding box edges including lines and paths."""
        hex_ = rgb_to_hex(self._current_edge_color)[0]
        return hex_to_name.get(hex_, hex_)

    @current_edge_color.setter
    def current_edge_color(self, edge_color):
        self._current_edge_color = transform_color(edge_color)
        if self._update_properties:
            for i in self.selected_data:
                self._data_view.update_edge_color(i, self._current_edge_color)
            self.events.edge_color()
            self._update_thumbnail()
        self.events.current_edge_color()

    @property
    def current_face_color(self):
        """str: color of bounding box faces."""
        hex_ = rgb_to_hex(self._current_face_color)[0]
        return hex_to_name.get(hex_, hex_)

    @current_face_color.setter
    def current_face_color(self, face_color):
        self._current_face_color = transform_color(face_color)
        if self._update_properties:
            for i in self.selected_data:
                self._data_view.update_face_color(i, self._current_face_color)
            self.events.face_color()
            self._update_thumbnail()
        self.events.current_face_color()

    @property
    def current_properties(self) -> Dict[str, np.ndarray]:
        """dict{str: np.ndarray(1,)}: properties for the next added shape."""
        return self._feature_table.currents()

    @current_properties.setter
    def current_properties(self, current_properties):
        update_indices = None
        if (
                self._update_properties
                and len(self.selected_data) > 0
                and self._mode in [Mode.SELECT, Mode.PAN_ZOOM]
        ):
            update_indices = list(self.selected_data)
        self._feature_table.set_currents(
            current_properties, update_indices=update_indices
        )
        if update_indices is not None:
            self.refresh_colors()
        self.events.current_properties()

    @property
    def nbounding_boxes(self):
        """int: Total number of bounding boxes."""
        return len(self._data_view.bounding_boxes)

    @property
    def _indices_view(self):
        return np.where(self._data_view._displayed)[0]

    @property
    def mode(self):
        """MODE: Interactive mode. The normal, default mode is PAN_ZOOM, which
        allows for normal interactivity with the canvas.

        The SELECT mode allows for entire bounding boxes to be selected, moved and
        resized.

        The ADD_BOUNDING_BOX mode allows for bounding boxes to be added.
        """
        return str(self._mode)

    @mode.setter
    def mode(self, mode):
        mode = Mode(mode)

        if not self.editable:
            mode = Mode.PAN_ZOOM

        if mode == self._mode:
            return
        old_mode = self._mode

        if old_mode in [Mode.SELECT, Mode.DIRECT]:
            self.mouse_drag_callbacks.remove(select)
            self.mouse_move_callbacks.remove(highlight)
        elif old_mode == Mode.ADD_BOUNDING_BOX:
            self.mouse_drag_callbacks.remove(add_bounding_box)

        if mode == Mode.PAN_ZOOM:
            self.cursor = 'standard'
            self.interactive = True
            self.help = trans._(
                'enter a selection mode to edit bounding box properties'
            )
        elif mode in [Mode.SELECT, Mode.DIRECT]:
            self.cursor = 'pointing'
            self.interactive = False
            self.help = trans._(
                'hold <space> to pan/zoom, press <{BACKSPACE}> to remove selected',
                BACKSPACE=BACKSPACE,
            )
            self.mouse_drag_callbacks.append(select)
            self.mouse_move_callbacks.append(highlight)

        elif mode == Mode.ADD_BOUNDING_BOX:
            self.cursor = 'cross'
            self.interactive = False
            self.help = trans._('hold <space> to pan/zoom')
            self.mouse_drag_callbacks.append(add_bounding_box)
        else:
            raise ValueError(
                trans._(
                    "Mode not recognized",
                    deferred=True,
                )
            )

        self._mode = mode

        draw_modes = [
            Mode.SELECT,
            Mode.DIRECT
        ]

        self.events.mode(mode=mode)

        # don't update thumbnail on mode changes
        with self.block_thumbnail_update():
            if not (mode in draw_modes and old_mode in draw_modes):
                # BoundingBoxLayer._finish_drawing() calls BoundingBoxLayer.refresh()
                self._finish_drawing()
            else:
                self.refresh()

    @property
    def z_index(self):
        """list of int: z_index for each bounding box."""
        return self._data_view.z_indices

    @z_index.setter
    def z_index(self, z_index):
        """Set z_index of bounding box using either int or list of int.

        When list of int is provided, must be of equal length to n bounding boxes.

        Parameters
        ----------
        z_index : int or list of int
            z-index of bounding boxes
        """
        if isinstance(z_index, list):
            if not len(z_index) == self.nbounding_boxes:
                raise ValueError(
                    trans._('Length of list does not match number of bounding boxes')
                )
            else:
                z_indices = z_index
        else:
            z_indices = [z_index for _ in range(self.nbounding_boxes)]

        for i, z_idx in enumerate(z_indices):
            self._data_view.update_z_index(i, z_idx)

    @property
    def edge_width(self):
        """list of float: edge width for each bounding box."""
        return self._data_view.edge_widths

    @edge_width.setter
    def edge_width(self, width):
        """Set edge width of bounding boxes using float or list of float.

        If list of float, must be of equal length to n bounding boxes

        Parameters
        ----------
        width : float or list of float
            width of all bounding boxes, or each bounding box if list
        """
        if isinstance(width, list):
            if not len(width) == self.nbounding_boxes:
                raise ValueError(
                    trans._('Length of list does not match number of bounding boxes')
                )
            else:
                widths = width
        else:
            widths = [width for _ in range(self.nbounding_boxes)]

        for i, width in enumerate(widths):
            self._data_view.update_edge_width(i, width)

    def _finish_drawing(self, event=None):
        """Reset properties used in bounding box drawing."""
        index = copy(self._moving_value[0])
        self._is_moving = False
        self.selected_data = set()
        self._drag_start = None
        self._drag_box = None
        self._is_selecting = False
        self._fixed_vertex = None
        self._value = (None, None)
        self._moving_value = (None, None)
        if self._is_creating is True and self._mode == Mode.ADD_PATH:
            vertices = self._data_view.bounding_boxes[index].data
            if len(vertices) <= 2:
                self._data_view.remove(index)
            else:
                self._data_view.edit(index, vertices[:-1])
        if self._is_creating is True and self._mode == Mode.ADD_POLYGON:
            vertices = self._data_view.bounding_boxes[index].data
            if len(vertices) <= 3:
                self._data_view.remove(index)
            else:
                self._data_view.edit(index, vertices[:-1])
        self._is_creating = False
        self._update_dims()

    def _set_color_cycle(self, color_cycle: np.ndarray, attribute: str):
        """Set the face_color_cycle or edge_color_cycle property

        Parameters
        ----------
        color_cycle : (N, 4) or (N, 1) array
            The value for setting edge or face_color_cycle
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        """
        transformed_color_cycle, transformed_colors = transform_color_cycle(
            color_cycle=color_cycle,
            elem_name=f'{attribute}_color_cycle',
            default="white",
        )
        setattr(self, f'_{attribute}_color_cycle_values', transformed_colors)
        setattr(self, f'_{attribute}_color_cycle', transformed_color_cycle)

        if self._update_properties is True:
            color_mode = getattr(self, f'_{attribute}_color_mode')
            if color_mode == ColorMode.CYCLE:
                self.refresh_colors(update_color_mapping=True)

    def _set_color(self, color, attribute: str):
        """Set the face_color or edge_color property

        Parameters
        ----------
        color : (N, 4) array or str
            The value for setting edge or face_color
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        """
        if self._is_color_mapped(color):
            if guess_continuous(self.properties[color]):
                setattr(self, f'_{attribute}_color_mode', ColorMode.COLORMAP)
            else:
                setattr(self, f'_{attribute}_color_mode', ColorMode.CYCLE)
            setattr(self, f'_{attribute}_color_property', color)
            self.refresh_colors()

        else:
            if len(self.data) > 0:
                transformed_color = transform_color_with_defaults(
                    num_entries=len(self.data),
                    colors=color,
                    elem_name="face_color",
                    default="white",
                )
                colors = normalize_and_broadcast_colors(
                    len(self.data), transformed_color
                )
            else:
                colors = np.empty((0, 4))

            setattr(self._data_view, f'{attribute}_color', colors)
            setattr(self, f'_{attribute}_color_mode', ColorMode.DIRECT)

            color_event = getattr(self.events, f'{attribute}_color')
            color_event()

    def _set_color_mode(
        self, color_mode: Union[ColorMode, str], attribute: str
    ):
        """Set the face_color_mode or edge_color_mode property

        Parameters
        ----------
        color_mode : str, ColorMode
            The value for setting edge or face_color_mode. If color_mode is a string,
            it should be one of: 'direct', 'cycle', or 'colormap'
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_colo_moder or 'face' for face_color_mode.
        """
        color_mode = ColorMode(color_mode)

        if color_mode == ColorMode.DIRECT:
            setattr(self, f'_{attribute}_color_mode', color_mode)
        elif color_mode in (ColorMode.CYCLE, ColorMode.COLORMAP):
            color_property = getattr(self, f'_{attribute}_color_property')
            if color_property == '':
                if self.properties:
                    new_color_property = next(iter(self.properties))
                    setattr(
                        self,
                        f'_{attribute}_color_property',
                        new_color_property,
                    )
                    warnings.warn(
                        trans._(
                            '_{attribute}_color_property was not set, setting to: {new_color_property}',
                            deferred=True,
                            attribute=attribute,
                            new_color_property=new_color_property,
                        )
                    )
                else:
                    raise ValueError(
                        trans._(
                            'There must be a valid BoundingBoxes.properties to use {color_mode}',
                            deferred=True,
                            color_mode=color_mode,
                        )
                    )

            # ColorMode.COLORMAP can only be applied to numeric properties
            color_property = getattr(self, f'_{attribute}_color_property')
            if (color_mode == ColorMode.COLORMAP) and not issubclass(
                self.properties[color_property].dtype.type, np.number
            ):
                raise TypeError(
                    trans._(
                        'selected property must be numeric to use ColorMode.COLORMAP',
                        deferred=True,
                    )
                )
            setattr(self, f'_{attribute}_color_mode', color_mode)
            self.refresh_colors()

    def refresh_text(self):
        """Refresh the text values.

        This is generally used if the properties were updated without changing the data
        """
        self.text.refresh_text(self.properties)

    def _is_color_mapped(self, color):
        """determines if the new color argument is for directly setting or cycle/colormap"""
        if isinstance(color, str):
            if color in self.properties:
                return True
            else:
                return False
        elif isinstance(color, (list, np.ndarray)):
            return False
        else:
            raise ValueError(
                trans._(
                    'face_color should be the name of a color, an array of colors, or the name of an property',
                    deferred=True,
                )
            )

    def _initialize_color(self, color, attribute: str, n_bounding_boxes: int):
        """Get the face/edge colors the BoundingBoxLayer layer will be initialized with

        Parameters
        ----------
        color : (N, 4) array or str
            The value for setting edge or face_color
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.

        Returns
        -------
        init_colors : (N, 4) array or str
            The calculated values for setting edge or face_color
        """
        if self._is_color_mapped(color):
            if guess_continuous(self.properties[color]):
                setattr(self, f'_{attribute}_color_mode', ColorMode.COLORMAP)
            else:
                setattr(self, f'_{attribute}_color_mode', ColorMode.CYCLE)
            setattr(self, f'_{attribute}_color_property', color)
            init_colors = self._map_color(
                attribute, update_color_mapping=False
            )

        else:
            if n_bounding_boxes > 0:
                transformed_color = transform_color_with_defaults(
                    num_entries=n_bounding_boxes,
                    colors=color,
                    elem_name="face_color",
                    default="white",
                )
                init_colors = normalize_and_broadcast_colors(
                    n_bounding_boxes, transformed_color
                )
            else:
                init_colors = np.empty((0, 4))

            setattr(self, f'_{attribute}_color_mode', ColorMode.DIRECT)

        return init_colors

    def _add_bounding_boxes(
        self,
        data,
        *,
        edge_width=None,
        edge_color=None,
        face_color=None,
        z_index=None
    ):
        """Add bounding boxes to the data view.

        Parameters
        ----------
        data : Array | List[Array]
            List of bounding box data, where each element is an (N, D) array of the
            N vertices of a bounding box in D dimensions. Can be an 3-dimensional array.
        edge_width : float | list
            thickness of lines and edges. If a list is supplied it must be the
            same length as the length of `data` and each element will be
            applied to each bounding box otherwise the same value will be used for all
            bounding boxes.
        edge_color : str | tuple | list
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements. If a list is supplied it must be the same length as
            the length of `data` and each element will be applied to each bounding box
            otherwise the same value will be used for all bounding boxes.
        face_color : str | tuple | list
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements. If a list is supplied it must be the same length as
            the length of `data` and each element will be applied to each bounding box
            otherwise the same value will be used for all bounding boxes.
        z_index : int | list
            Specifier of z order priority. Bounding boxes with higher z order are
            displayed ontop of others. If a list is supplied it must be the
            same length as the length of `data` and each element will be
            applied to each bounding box otherwise the same value will be used for all
            bounding boxes.
        """
        if edge_width is None:
            edge_width = self.current_edge_width
        if edge_color is None:
            edge_color = self._current_edge_color
        if face_color is None:
            face_color = self._current_face_color
        if self._data_view is not None:
            z_index = z_index or max(self._data_view._z_index, default=-1) + 1
        else:
            z_index = z_index or 0

        if len(data) > 0:
            if np.array(data[0]).ndim == 1:
                # If a single array for a bounding box has been passed turn into list
                data = [data]

            # transform the colors
            transformed_ec = transform_color_with_defaults(
                num_entries=len(data),
                colors=edge_color,
                elem_name="edge_color",
                default="white",
            )
            transformed_edge_color = normalize_and_broadcast_colors(
                len(data), transformed_ec
            )
            transformed_fc = transform_color_with_defaults(
                num_entries=len(data),
                colors=face_color,
                elem_name="face_color",
                default="white",
            )
            transformed_face_color = normalize_and_broadcast_colors(
                len(data), transformed_fc
            )

            # Turn input arguments into iterables
            bounding_box_inputs = zip(
                data,
                ensure_iterable(edge_width),
                transformed_edge_color,
                transformed_face_color,
                ensure_iterable(z_index),
            )

            self._add_bounding_boxes_to_view(bounding_box_inputs, self._data_view)

        self._display_order_stored = copy(self._dims_order)
        self._ndisplay_stored = copy(self._ndisplay)
        self._update_dims()

    def _add_bounding_boxes_to_view(self, bounding_box_inputs, data_view):
        """Build new bounding boxes and add them to the _data_view"""
        for d, ew, ec, fc, z in bounding_box_inputs:

            bounding_box = BoundingBox(
                d,
                edge_width=ew,
                z_index=z,
                dims_order=self._dims_order,
                ndisplay=self._ndisplay,
            )

            # Add bounding box
            data_view.add(bounding_box, edge_color=ec, face_color=fc, z_refresh=False)
        data_view._update_z_order()

    def add(
        self,
        data,
        *,
        edge_width=None,
        edge_color=None,
        face_color=None,
        z_index=None,
    ):
        """Add bounding boxes to the current layer.

        Parameters
        ----------
        data : Array | List[Array]
            List of bounding box data, where each element is an (N, D) array of the
            N vertices of a bounding box in D dimensions. Can be an 3-dimensional array.
        edge_width : float | list
            thickness of lines and edges. If a list is supplied it must be the
            same length as the length of `data` and each element will be
            applied to each bounding box otherwise the same value will be used for all
            bounding boxes.
        edge_color : str | tuple | list
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements. If a list is supplied it must be the same length as
            the length of `data` and each element will be applied to each bounding box
            otherwise the same value will be used for all bounding boxes.
        face_color : str | tuple | list
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements. If a list is supplied it must be the same length as
            the length of `data` and each element will be applied to each bounding box
            otherwise the same value will be used for all bounding boxes.
        z_index : int | list
            Specifier of z order priority. Bounding boxes with higher z order are
            displayed ontop of others. If a list is supplied it must be the
            same length as the length of `data` and each element will be
            applied to each bounding box otherwise the same value will be used for all
            bounding boxes.
        """

        if edge_width is None:
            edge_width = self.current_edge_width

        n_new_bounding_boxes = len(data) if type(data) == list or data.ndim > 2 else 1
        if edge_color is None:
            edge_color = self._get_new_bounding_box_color(
                n_new_bounding_boxes, attribute='edge'
            )
        if face_color is None:
            face_color = self._get_new_bounding_box_color(
                n_new_bounding_boxes, attribute='face'
            )
        if self._data_view is not None:
            z_index = z_index or max(self._data_view._z_index, default=-1) + 1
        else:
            z_index = z_index or 0

        if n_new_bounding_boxes > 0:
            if len(self.properties) > 0:
                first_prop_key = next(iter(self.properties))
                n_prop_values = len(self.properties[first_prop_key])
            else:
                n_prop_values = 0
            total_bounding_boxes = n_new_bounding_boxes + self.nbounding_boxes
            self._feature_table.resize(total_bounding_boxes)
            if total_bounding_boxes > n_prop_values:
                n_props_to_add = total_bounding_boxes - n_prop_values
                self.text.add(self.current_properties, n_props_to_add)
            if total_bounding_boxes < n_prop_values:
                for k in self.properties:
                    self.properties[k] = self.properties[k][:total_bounding_boxes]
                n_props_to_remove = n_prop_values - total_bounding_boxes
                indices_to_remove = np.arange(n_prop_values)[
                    -n_props_to_remove:
                ]
                self.text.remove(indices_to_remove)

            self._add_bounding_boxes(
                data,
                edge_width=edge_width,
                edge_color=edge_color,
                face_color=face_color,
                z_index=z_index,
            )

    def _init_bounding_boxes(
        self,
        data,
        *,
        edge_width=None,
        edge_color=None,
        edge_color_cycle,
        edge_colormap,
        edge_contrast_limits,
        face_color=None,
        face_color_cycle,
        face_colormap,
        face_contrast_limits,
        z_index=None,
    ):
        """Add bounding boxes to the data view.

        Parameters
        ----------
        data : Array | List[Array]
            List of bounding box data, where each element is an (N, D) array of the
            N vertices of a bounding box in D dimensions. Can be an 3-dimensional array.
        edge_width : float | list
            thickness of lines and edges. If a list is supplied it must be the
            same length as the length of `data` and each element will be
            applied to each bounding box otherwise the same value will be used for all
            bounding boxes.
        edge_color : str | tuple | list
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements. If a list is supplied it must be the same length as
            the length of `data` and each element will be applied to each bounding box
            otherwise the same value will be used for all bounding boxes.
        face_color : str | tuple | list
            If string can be any color name recognized by vispy or hex value if
            starting with `#`. If array-like must be 1-dimensional array with 3
            or 4 elements. If a list is supplied it must be the same length as
            the length of `data` and each element will be applied to each bounding box
            otherwise the same value will be used for all bounding boxes.
        z_index : int | list
            Specifier of z order priority. Bounding boxes with higher z order are
            displayed ontop of others. If a list is supplied it must be the
            same length as the length of `data` and each element will be
            applied to each bounding box otherwise the same value will be used for all
            bounding boxes.
        """

        n_bounding_boxes = len(data)
        with self.block_update_properties():
            self._edge_color_property = ''
            self.edge_color_cycle_map = {}
            self.edge_colormap = edge_colormap
            self._edge_contrast_limits = edge_contrast_limits
            if edge_color_cycle is None:
                edge_color_cycle = deepcopy(DEFAULT_COLOR_CYCLE)
            self.edge_color_cycle = edge_color_cycle
            edge_color = self._initialize_color(
                edge_color, attribute='edge', n_bounding_boxes=n_bounding_boxes
            )

            self._face_color_property = ''
            self.face_color_cycle_map = {}
            self.face_colormap = face_colormap
            self._face_contrast_limits = face_contrast_limits
            if face_color_cycle is None:
                face_color_cycle = deepcopy(DEFAULT_COLOR_CYCLE)
            self.face_color_cycle = face_color_cycle
            face_color = self._initialize_color(
                face_color, attribute='face', n_bounding_boxes=n_bounding_boxes
            )

        with self.block_thumbnail_update():
            self._add_bounding_boxes(
                data,
                edge_width=edge_width,
                edge_color=edge_color,
                face_color=face_color,
                z_index=z_index
            )
            self._data_view._update_z_order()
            self.refresh_colors()

    def refresh_colors(self, update_color_mapping: bool = False):
        """Calculate and update face and edge colors if using a cycle or color map

        Parameters
        ----------
        update_color_mapping : bool
            If set to True, the function will recalculate the color cycle map
            or colormap (whichever is being used). If set to False, the function
            will use the current color cycle map or color map. For example, if you
            are adding/modifying bounding boxes and want them to be colored with the same
            mapping as the other bounding boxes (i.e., the new bounding boxes shouldn't affect
            the color cycle map or colormap), set update_color_mapping=False.
            Default value is False.
        """

        self._refresh_color('face', update_color_mapping)
        self._refresh_color('edge', update_color_mapping)

    def _refresh_color(
        self, attribute: str, update_color_mapping: bool = False
    ):
        """Calculate and update face or edge colors if using a cycle or color map

        Parameters
        ----------
        attribute : str  in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        update_color_mapping : bool
            If set to True, the function will recalculate the color cycle map
            or colormap (whichever is being used). If set to False, the function
            will use the current color cycle map or color map. For example, if you
            are adding/modifying bounding boxes and want them to be colored with the same
            mapping as the other bounding boxes (i.e., the new bounding boxes shouldn't affect
            the color cycle map or colormap), set update_color_mapping=False.
            Default value is False.
        """
        if self._update_properties:
            color_mode = getattr(self, f'_{attribute}_color_mode')
            if color_mode in [ColorMode.CYCLE, ColorMode.COLORMAP]:
                colors = self._map_color(attribute, update_color_mapping)
                setattr(self._data_view, f'{attribute}_color', colors)

                color_event = getattr(self.events, f'{attribute}_color')
                color_event()

    def _get_new_bounding_box_color(self, adding: int, attribute: str):
        """Get the color for the bounding box(es) to be added.

        Parameters
        ----------
        adding : int
            the number of bounding boxes that were added
            (and thus the number of color entries to add)
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color_mode or 'face' for face_color_mode.

        Returns
        -------
        new_colors : (N, 4) array
            (Nx4) RGBA array of colors for the N new bounding boxes
        """
        color_mode = getattr(self, f'_{attribute}_color_mode')
        if color_mode == ColorMode.DIRECT:
            current_face_color = getattr(self, f'_current_{attribute}_color')
            new_colors = np.tile(current_face_color, (adding, 1))
        elif color_mode == ColorMode.CYCLE:
            property_name = getattr(self, f'_{attribute}_color_property')
            color_property_value = self.current_properties[property_name][0]

            # check if the new color property is in the cycle map
            # and add it if it is not
            color_cycle_map = getattr(self, f'{attribute}_color_cycle_map')
            color_cycle_keys = [*color_cycle_map]
            if color_property_value not in color_cycle_keys:
                color_cycle = getattr(self, f'_{attribute}_color_cycle')
                color_cycle_map[color_property_value] = np.squeeze(
                    transform_color(next(color_cycle))
                )

                setattr(self, f'{attribute}_color_cycle_map', color_cycle_map)

            new_colors = np.tile(
                color_cycle_map[color_property_value], (adding, 1)
            )
        elif color_mode == ColorMode.COLORMAP:
            property_name = getattr(self, f'_{attribute}_color_property')
            color_property_value = self.current_properties[property_name][0]
            colormap = getattr(self, f'{attribute}_colormap')
            contrast_limits = getattr(self, f'_{attribute}_contrast_limits')

            fc, _ = map_property(
                prop=color_property_value,
                colormap=colormap,
                contrast_limits=contrast_limits,
            )
            new_colors = np.tile(fc, (adding, 1))

        return new_colors

    def _map_color(self, attribute: str, update_color_mapping: bool = False):
        """Calculate the mapping for face or edge colors if using a cycle or color map

        Parameters
        ----------
        attribute : str  in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        update_color_mapping : bool
            If set to True, the function will recalculate the color cycle map
            or colormap (whichever is being used). If set to False, the function
            will use the current color cycle map or color map. For example, if you
            are adding/modifying bounding boxes and want them to be colored with the same
            mapping as the other bounding boxes (i.e., the new bounding boxes shouldn't affect
            the color cycle map or colormap), set update_color_mapping=False.
            Default value is False.

        Returns
        -------
        colors : (N, 4) array or str
            The calculated values for setting edge or face_color
        """
        color_mode = getattr(self, f'_{attribute}_color_mode')
        if color_mode == ColorMode.CYCLE:
            color_property = getattr(self, f'_{attribute}_color_property')
            color_properties = self.properties[color_property]
            if update_color_mapping:
                color_cycle = getattr(self, f'_{attribute}_color_cycle')
                color_cycle_map = {
                    k: np.squeeze(transform_color(c))
                    for k, c in zip(np.unique(color_properties), color_cycle)
                }
                setattr(self, f'{attribute}_color_cycle_map', color_cycle_map)

            else:
                # add properties if they are not in the colormap
                # and update_color_mapping==False
                color_cycle_map = getattr(self, f'{attribute}_color_cycle_map')
                color_cycle_keys = [*color_cycle_map]
                props_in_map = np.in1d(color_properties, color_cycle_keys)
                if not np.all(props_in_map):
                    props_to_add = np.unique(
                        color_properties[np.logical_not(props_in_map)]
                    )
                    color_cycle = getattr(self, f'_{attribute}_color_cycle')
                    for prop in props_to_add:
                        color_cycle_map[prop] = np.squeeze(
                            transform_color(next(color_cycle))
                        )
                    setattr(
                        self,
                        f'{attribute}_color_cycle_map',
                        color_cycle_map,
                    )
            colors = np.array([color_cycle_map[x] for x in color_properties])
            if len(colors) == 0:
                colors = np.empty((0, 4))

        elif color_mode == ColorMode.COLORMAP:
            color_property = getattr(self, f'_{attribute}_color_property')
            color_properties = self.properties[color_property]
            if len(color_properties) > 0:
                contrast_limits = getattr(self, f'{attribute}_contrast_limits')
                colormap = getattr(self, f'{attribute}_colormap')
                if update_color_mapping or contrast_limits is None:

                    colors, contrast_limits = map_property(
                        prop=color_properties, colormap=colormap
                    )
                    setattr(
                        self,
                        f'{attribute}_contrast_limits',
                        contrast_limits,
                    )
                else:

                    colors, _ = map_property(
                        prop=color_properties,
                        colormap=colormap,
                        contrast_limits=contrast_limits,
                    )
            else:
                colors = np.empty((0, 4))

        return colors

    @contextmanager
    def block_thumbnail_update(self):
        """Use this context manager to block thumbnail updates"""
        self._allow_thumbnail_update = False
        yield
        self._allow_thumbnail_update = True

    def _set_view_slice(self):
        """Set the view given the slicing indices."""
        if not self._ndisplay == self._ndisplay_stored:
            self.selected_data = set()
            self._data_view.ndisplay = min(self.ndim, self._ndisplay)
            self._ndisplay_stored = copy(self._ndisplay)
            self._clipboard = {}

        if not self._dims_order == self._display_order_stored:
            self.selected_data = set()
            self._data_view.update_dims_order(self._dims_order)
            self._display_order_stored = copy(self._dims_order)
            # Clear clipboard if dimensions swap
            self._clipboard = {}

        slice_key = np.array(self._slice_indices)[
            list(self._dims_not_displayed)
        ]
        if not np.all(slice_key == self._data_view.slice_key):
            self.selected_data = set()
        self._data_view.slice_key = slice_key

    def _update_thumbnail(self, event=None):
        """Update thumbnail with current bounding boxes and colors."""

        # don't update the thumbnail if dragging a bounding box
        if self._is_moving is False and self._allow_thumbnail_update is True:
            # calculate min vals for the vertices and pad with 0.5
            # the offset is needed to ensure that the top left corner of the bounding boxes
            # corresponds to the top left corner of the thumbnail
            de = self._extent_data
            offset = np.array([de[0, d] for d in self._dims_displayed]) + 0.5
            # calculate range of values for the vertices and pad with 1
            # padding ensures the entire bounding box can be represented in the thumbnail
            # without getting clipped
            bounding_box = np.ceil(
                [de[1, d] - de[0, d] + 1 for d in self._dims_displayed]
            ).astype(int)
            zoom_factor = np.divide(
                self._thumbnail_shape[:2], bounding_box[-2:]
            ).min()

            colormapped = self._data_view.to_colors(
                colors_shape=self._thumbnail_shape[:2],
                zoom_factor=zoom_factor,
                offset=offset[-2:],
                max_bounding_boxes=self._max_bounding_boxes_thumbnail,
            )

            self.thumbnail = colormapped

    def remove_selected(self):
        """Remove any selected bounding boxes."""
        index = list(self.selected_data)
        to_remove = sorted(index, reverse=True)
        for ind in to_remove:
            self._data_view.remove(ind)

        if len(index) > 0:
            self._feature_table.remove(index)
            self.text.remove(index)
            self._data_view._edge_color = np.delete(
                self._data_view._edge_color, index, axis=0
            )
            self._data_view._face_color = np.delete(
                self._data_view._face_color, index, axis=0
            )
        self.selected_data = set()
        self._finish_drawing()

    def interaction_box(self, index):
        """Create the interaction box around a bounding box or list of bounding boxes.
        If a single index is passed then the boudning box will be inherited
        from that bounding boxes interaction box. If list of indices is passed it will
        be computed directly.

        Parameters
        ----------
        index : int | list
            Index of a single bounding box, or a list of bounding boxes around which to
            construct the interaction box

        Returns
        -------
        box : np.ndarray
            10x2 array of vertices of the interaction box. The first 8 points
            are the corners and midpoints of the box in clockwise order
            starting in the upper-left corner. The 9th point is the center of
            the box, and the last point is the location of the rotation handle
            that can be used to rotate the box
        """
        if isinstance(index, (list, np.ndarray, set)):
            if len(index) == 0:
                box = None
            elif len(index) == 1:
                box = copy(self._data_view.bounding_boxes[list(index)[0]]._box)
            else:
                indices = np.isin(self._data_view.displayed_index, list(index))
                box = create_box(self._data_view.displayed_vertices[indices])
        else:
            box = copy(self._data_view.bounding_boxes[index]._box)

        if box is not None:
            rot = box[Box.TOP_CENTER]
            length_box = np.linalg.norm(
                box[Box.BOTTOM_LEFT] - box[Box.TOP_LEFT]
            )
            if length_box > 0:
                r = self._rotation_handle_length * self.scale_factor
                rot = (
                    rot
                    - r
                    * (box[Box.BOTTOM_LEFT] - box[Box.TOP_LEFT])
                    / length_box
                )
            box = np.append(box, [rot], axis=0)

        return box

    def _scale_box(self, scale, center=[0, 0]):
        """Perform a scaling on the selected box.

        Parameters
        ----------
        scale : float, list
            scalar or list specifying rescaling of bounding box.
        center : list
            coordinates of center of rotation.
        """
        if not isinstance(scale, (list, np.ndarray)):
            scale = [scale, scale]
        box = self._selected_box - center
        box = np.array(box * scale)
        # if not np.all(box[Box.TOP_CENTER] == box[Box.HANDLE]):
        #     r = self._rotation_handle_length * self.scale_factor
        #     handle_vec = box[Box.HANDLE] - box[Box.TOP_CENTER]
        #     cur_len = np.linalg.norm(handle_vec)
        #     box[Box.HANDLE] = box[Box.TOP_CENTER] + r * handle_vec / cur_len
        self._selected_box = box + center

    def _transform_box(self, transform, center=[0, 0]):
        """Perform a linear transformation on the selected box.

        Parameters
        ----------
        transform : np.ndarray
            2x2 array specifying linear transform.
        center : list
            coordinates of center of rotation.
        """
        box = self._selected_box - center
        box = box @ transform.T
        if not np.all(box[Box.TOP_CENTER] == box[Box.HANDLE]):
            r = self._rotation_handle_length * self.scale_factor
            handle_vec = box[Box.HANDLE] - box[Box.TOP_CENTER]
            cur_len = np.linalg.norm(handle_vec)
            box[Box.HANDLE] = box[Box.TOP_CENTER] + r * handle_vec / cur_len
        self._selected_box = box + center

    def _get_value(self, position):
        """Value of the data at a position in data coordinates.

        Parameters
        ----------
        position : tuple
            Position in data coordinates.

        Returns
        -------
        bounding box : int | None
            Index of bounding box if any that is at the coordinates. Returns `None`
            if no bounding box is found.
        vertex : int | None
            Index of vertex if any that is at the coordinates. Returns `None`
            if no vertex is found.
        """
        if self._ndisplay == 3:
            return (None, None)

        if self._is_moving:
            return self._moving_value

        coord = [position[i] for i in self._dims_displayed]

        # Check selected bounding boxes
        value = None
        selected_index = list(self.selected_data)
        if len(selected_index) > 0:
            if self._mode == Mode.SELECT:
                # Check if inside vertex of interaction box or rotation handle
                box = self._selected_box[Box.WITHOUT_HANDLE]
                distances = abs(box - coord)

                # Get the vertex sizes
                sizes = self._vertex_size * self.scale_factor / 2

                # Check if any matching vertices
                matches = np.all(distances <= sizes, axis=1).nonzero()
                if len(matches[0]) > 0:
                    value = (selected_index[0], matches[0][-1])
            elif self._mode == Mode.DIRECT:
                # Check if inside vertex of bounding box
                inds = np.isin(self._data_view.displayed_index, selected_index)
                vertices = self._data_view.displayed_vertices[inds]
                distances = abs(vertices - coord)

                # Get the vertex sizes
                sizes = self._vertex_size * self.scale_factor / 2

                # Check if any matching vertices
                matches = np.all(distances <= sizes, axis=1).nonzero()[0]
                if len(matches) > 0:
                    index = inds.nonzero()[0][matches[-1]]
                    bounding_box = self._data_view.displayed_index[index]
                    vals, idx = np.unique(
                        self._data_view.displayed_index, return_index=True
                    )
                    bounding_box_in_list = list(vals).index(bounding_box)
                    value = (bounding_box, index - idx[bounding_box_in_list])

        if value is None:
            # Check if mouse inside bounding box
            bounding_box = self._data_view.inside(coord)
            value = (bounding_box, None)

        return value

    def move_to_front(self):
        """Moves selected objects to be displayed in front of all others."""
        if len(self.selected_data) == 0:
            return
        new_z_index = max(self._data_view._z_index) + 1
        for index in self.selected_data:
            self._data_view.update_z_index(index, new_z_index)
        self.refresh()

    def move_to_back(self):
        """Moves selected objects to be displayed behind all others."""
        if len(self.selected_data) == 0:
            return
        new_z_index = min(self._data_view._z_index) - 1
        for index in self.selected_data:
            self._data_view.update_z_index(index, new_z_index)
        self.refresh()

    def _copy_data(self):
        """Copy selected shapes to clipboard."""
        if len(self.selected_data) > 0:
            index = list(self.selected_data)
            self._clipboard = {
                'data': [
                    deepcopy(self._data_view.bounding_boxes[i])
                    for i in self._selected_data
                ],
                'edge_color': deepcopy(self._data_view._edge_color[index]),
                'face_color': deepcopy(self._data_view._face_color[index]),
                'features': deepcopy(self.features.iloc[index]),
                'indices': self._slice_indices,
            }
            if len(self.text.values) == 0:
                self._clipboard['text'] = np.empty(0)
            else:
                self._clipboard['text'] = deepcopy(self.text.values[index])
        else:
            self._clipboard = {}

    def _paste_data(self):
        """Paste any shapes from clipboard and then selects them."""
        cur_bboxes = self.nbounding_boxes
        if len(self._clipboard.keys()) > 0:
            # Calculate offset based on dimension shifts
            offset = [
                self._slice_indices[i] - self._clipboard['indices'][i]
                for i in self._dims_not_displayed
            ]

            self._feature_table.append(self._clipboard['features'])

            # Add new shape data
            for i, bb in enumerate(self._clipboard['data']):
                bbox = deepcopy(bb)
                data = copy(bbox.data)
                data[:, self._dims_not_displayed] = data[
                    :, self._dims_not_displayed
                ] + np.array(offset)
                bbox.data = data
                face_color = self._clipboard['face_color'][i]
                edge_color = self._clipboard['edge_color'][i]
                self._data_view.add(
                    bbox, face_color=face_color, edge_color=edge_color
                )

            if len(self._clipboard['text']) > 0:
                self.text.values = np.concatenate(
                    (self.text.values, self._clipboard['text']), axis=0
                )

            self.selected_data = set(
                range(cur_bboxes, cur_bboxes + len(self._clipboard['data']))
            )

            self.move_to_front()

    def _set_highlight(self, force=False):
        """Render highlights of bounding boxes.

        Includes boundaries, vertices, interaction boxes, and the drag
        selection box when appropriate.

        Parameters
        ----------
        force : bool
            Bool that forces a redraw to occur when `True`
        """
        # Check if any bounding box or vertex ids have changed since last call
        if (
            self.selected_data == self._selected_data_stored
            and np.all(self._value == self._value_stored)
            and np.all(self._drag_box == self._drag_box_stored)
        ) and not force:
            return
        self._selected_data_stored = copy(self.selected_data)
        self._value_stored = copy(self._value)
        self._drag_box_stored = copy(self._drag_box)
        self.events.highlight()

    def _get_ndim(self):
        """Determine number of dimensions of the layer."""
        if self.nbounding_boxes == 0:
            ndim = self.ndim
        else:
            ndim = self.data[0].shape[1]
        return ndim

    def _get_state(self):
        """Get dictionary of layer state.

        Returns
        -------
        state : dict
            Dictionary of layer state.
        """
        state = self._get_base_state()
        state.update(
            {
                'ndim': self.ndim,
                'properties': self.properties,
                'text': self.text.dict(),
                'opacity': self.opacity,
                'z_index': self.z_index,
                'edge_width': self.edge_width,
                'face_color': self.face_color,
                'face_color_cycle': self.face_color_cycle,
                'face_colormap': self.face_colormap.name,
                'face_contrast_limits': self.face_contrast_limits,
                'edge_color': self.edge_color,
                'edge_color_cycle': self.edge_color_cycle,
                'edge_colormap': self.edge_colormap.name,
                'edge_contrast_limits': self.edge_contrast_limits,
                'data': self.data,
                'features': self.features,
            }
        )
        return state

    @property
    def _view_text(self) -> np.ndarray:
        """Get the values of the text elements in view

        Returns
        -------
        text : (N x 1) np.ndarray
            Array of text strings for the N text elements in view
        """
        return self.text.view_text(self._indices_view)

    @property
    def _view_text_coords(self) -> np.ndarray:
        """Get the coordinates of the text elements in view

        Returns
        -------
        text_coords : (N x D) np.ndarray
            Array of coordindates for the N text elements in view
        """
        # get the coordinates of the vertices for the bounding boxes in view
        in_view_bounding_boxes_coords = [
            self._data_view.data[i] for i in self._indices_view
        ]

        # get the coordinates for the dimensions being displayed
        sliced_in_view_coords = [
            position[:, self._dims_displayed]
            for position in in_view_bounding_boxes_coords
        ]

        return self.text.compute_text_coords(
            sliced_in_view_coords, self._ndisplay
        )

    def _outline_bounding_boxes(self):
        """Find outlines of any selected or hovered bounding boxes.

        Returns
        -------
        vertices : None | np.ndarray
            Nx2 array of any vertices of outline or None
        triangles : None | np.ndarray
            Mx3 array of any indices of vertices for triangles of outline or
            None
        """
        if self._value is not None and (
            self._value[0] is not None or len(self.selected_data) > 0
        ):
            if len(self.selected_data) > 0:
                index = list(self.selected_data)
                if self._value[0] is not None:
                    if self._value[0] in index:
                        pass
                    else:
                        index.append(self._value[0])
                index.sort()
            else:
                index = self._value[0]

            centers, offsets, triangles = self._data_view.outline(index)
            vertices = centers + (
                self.scale_factor * self._highlight_width * offsets
            )
            vertices = vertices[:, ::-1]
        else:
            vertices = None
            triangles = None

        return vertices, triangles

    def _compute_vertices_and_box(self):
        """Compute location of highlight vertices and box for rendering.

        Returns
        -------
        vertices : np.ndarray
            Nx2 array of any vertices to be rendered as Markers
        face_color : str
            String of the face color of the Markers
        edge_color : str
            String of the edge color of the Markers and Line for the box
        pos : np.ndarray
            Nx2 array of vertices of the box that will be rendered using a
            Vispy Line
        width : float
            Width of the box edge
        """
        if len(self.selected_data) > 0:
            if self._mode == Mode.SELECT:
                # If in select mode just show the interaction boudning box
                # including its vertices and the rotation handle
                box = self._selected_box[Box.WITHOUT_HANDLE]
                if self._value[0] is None:
                    face_color = 'white'
                elif self._value[1] is None:
                    face_color = 'white'
                else:
                    face_color = self._highlight_color
                edge_color = self._highlight_color
                vertices = box[:, ::-1]
                # Use a subset of the vertices of the interaction_box to plot
                # the line around the edge
                pos = box[Box.LINE_HANDLE][:, ::-1]
                width = 1.5
            elif self._mode in (
                [
                    Mode.DIRECT,
                    Mode.ADD_BOUNDING_BOX
                ]
            ):
                # If in one of these mode show the vertices of the bounding box itself
                inds = np.isin(
                    self._data_view.displayed_index, list(self.selected_data)
                )
                vertices = self._data_view.displayed_vertices[inds][:, ::-1]
                # If currently adding path don't show box over last vertex

                if self._value[0] is None:
                    face_color = 'white'
                elif self._value[1] is None:
                    face_color = 'white'
                else:
                    face_color = self._highlight_color
                edge_color = self._highlight_color
                pos = None
                width = 0
            else:
                # Otherwise show nothing
                vertices = np.empty((0, 2))
                face_color = 'white'
                edge_color = 'white'
                pos = None
                width = 0
        elif self._is_selecting:
            # If currently dragging a selection box just show an outline of
            # that box
            vertices = np.empty((0, 2))
            edge_color = self._highlight_color
            face_color = 'white'
            box = create_box(self._drag_box)
            width = 1.5
            # Use a subset of the vertices of the interaction_box to plot
            # the line around the edge
            pos = box[Box.LINE][:, ::-1]
        else:
            # Otherwise show nothing
            vertices = np.empty((0, 2))
            face_color = 'white'
            edge_color = 'white'
            pos = None
            width = 0

        return vertices, face_color, edge_color, pos, width

    def _set_editable(self, editable=None):
        if editable is None:
            if self._ndisplay == 3:
                self.editable = False
            else:
                self.editable = True

        if not self.editable:
            self.mode = Mode.PAN_ZOOM


# This is an ugly solution to register every component correctly
register_layer_control(BoundingBoxLayer)
register_layer_visual(BoundingBoxLayer)
# register_bounding_boxes_actions(BoundingBoxLayer)
import sys
napari.layers.__dict__["BoundingBoxLayer"] = BoundingBoxLayer
napari.layers.__dict__["bounding_boxes"] = sys.modules[__name__]



from qtpy.QtCore import QObject, QEvent
from qtpy.QtWidgets import QWidget, QVBoxLayout, QCheckBox, QTabWidget, QLabel
from napari import Viewer
from napari.layers import Labels, Layer
import numpy as np
from skimage.draw import draw
from scipy.ndimage import binary_fill_holes
import math

from .interpolation_widget import InterpolationWidget


class AnnotatorWidget(QWidget):
    def __init__(self, viewer: Viewer):
        print("AnnotatorWidget init")
        super().__init__()
        layout = QVBoxLayout()
        self.fill_objects_checkbox = QCheckBox("autofill objects")
        self.fill_objects_checkbox.setChecked(True)
        self._active_labels_layer = None
        self.viewer = viewer
        self.viewer.layers.selection.events.connect(self.on_layer_selection_change)
        self.fill_objects_checkbox.clicked.connect(self.set_fill_objects)
        layout.addWidget(self.fill_objects_checkbox)
        tabs_widget = QTabWidget()
        annotator_j_widget = QLabel("annotatorJ")
        tabs_widget.addTab(annotator_j_widget, "annotatorJ")
        minimal_surface_widget = QLabel("Minimal surface ")
        tabs_widget.addTab(minimal_surface_widget, "Minimal surface")
        interpolation_widget = InterpolationWidget(viewer)
        tabs_widget.addTab(interpolation_widget, "Interpolation")
        layout.addWidget(tabs_widget)
        self.setLayout(layout)
        self.installEventFilter(self)
        layout.addStretch()
        self.set_fill_objects(True)
        self.on_layer_selection_change()

    def eventFilter(self, source: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Hide:
            self.disconnect_all()
            return True
        elif event.type() == QEvent.Show:
            self.connect_all()
            return True
        return super().eventFilter(source, event)

    @property
    def fill_objects(self):
        return self.fill_objects_checkbox.isChecked()

    @property
    def active_labels_layer(self):
        return self._active_labels_layer

    @active_labels_layer.setter
    def active_labels_layer(self, labels_layer):
        if not isinstance(labels_layer, Labels) and labels_layer is not None:
            return
        self._active_labels_layer = labels_layer
        self.set_fill_objects(self.fill_objects)

    def set_fill_objects(self, state):
        if state and self.isVisible() and self.active_labels_layer is not None and\
                self.fill_holes not in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.append(self.fill_holes)
        elif not state and self.active_labels_layer is not None and\
                self.fill_holes in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.remove(self.fill_holes)

    def disconnect_all(self):
        if self.active_labels_layer is not None and self.fill_holes in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.remove(self.fill_holes)

    def connect_all(self):
        if self.fill_objects and self.active_labels_layer is not None and self.fill_holes not in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.append(self.fill_holes)

    def on_layer_selection_change(self, event=None):
        if event is None or event.type == "changed":
            print("change")
            active_layer = self.viewer.layers.selection.active
            if isinstance(active_layer, Labels):
                self.active_labels_layer = active_layer
            else:
                self.active_labels_layer = None

    def draw_line(self, x1, y1, x2, y2, brush_size, output):
        line_x, line_y = draw.line(x1, y1, x2, y2)
        for x, y in zip(line_x, line_y):
            cx, cy = draw.disk((x, y), math.ceil(brush_size/2+0.1))
            cx = np.clip(cx, 0, output.shape[0] - 1)
            cy = np.clip(cy, 0, output.shape[1] - 1)
            output[cx, cy] = True

    def fill_holes(self, layer: Layer, event):
        if layer.mode != "paint" or layer._ndisplay != 2:
            return
        coordinates = layer.world_to_data(event.position)
        coordinates = tuple(max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
        image_coords = tuple(coordinates[i] for i in range(len(coordinates)) if i in self.viewer.dims.displayed)
        slice_dims = tuple(coordinates[i] if i in self.viewer.dims.not_displayed else slice(None) for i in range(len(coordinates)))
        current_draw = np.zeros_like(layer.data[slice_dims], np.bool)
        start_x, start_y = prev_x, prev_y = image_coords
        cx, cy = draw.disk((start_x, start_y), layer.brush_size/2)
        cx = np.clip(cx, 0, current_draw.shape[0] - 1)
        cy = np.clip(cy, 0, current_draw.shape[1] - 1)
        current_draw[cx, cy] = True
        yield
        while event.type == 'mouse_move':

            coordinates = layer.world_to_data(event.position)
            coordinates = tuple(max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
            image_coords = tuple(coordinates[i] for i in range(len(coordinates)) if i in self.viewer.dims.displayed)
            self.draw_line(prev_x, prev_y, image_coords[-2], image_coords[-1], layer.brush_size, current_draw)
            prev_x, prev_y = image_coords
            yield
        # s = np.asarray([[0, 1, 0],
        #                 [1, 1, 1],
        #                 [0, 1, 0]])
        s = None
        coordinates = layer.world_to_data(event.position)
        coordinates = tuple(
            max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
        image_coords = tuple(coordinates[i] for i in range(len(coordinates)) if i in self.viewer.dims.displayed)
        prev_x, prev_y = image_coords
        self.draw_line(prev_x, prev_y, start_x, start_y, layer.brush_size, current_draw)
        cx, cy = draw.disk((prev_x, prev_y), layer.brush_size/2)
        cx = np.clip(cx, 0, current_draw.shape[0] - 1)
        cy = np.clip(cy, 0, current_draw.shape[1] - 1)
        current_draw[cx, cy] = True
        binary_fill_holes(current_draw, output=current_draw, structure=s)
        layer.data[slice_dims][current_draw] = layer.selected_label
        layer.refresh()



# A copy of napari.layers.shapes._shapes_constants
import sys
from enum import auto

from napari.utils.misc import StringEnum


class Mode(StringEnum):
    """Mode: Interactive mode. The normal, default mode is PAN_ZOOM, which
    allows for normal interactivity with the canvas.

    The SELECT mode allows for bounding boxes to be selected, moved and
    resized.

    The ADD_BOUNDING_BOX mode allows for bounding boxes to be added.

    The DIRECT mode is unused, and will be removed
    """

    PAN_ZOOM = auto()
    SELECT = auto()
    DIRECT = auto()
    ADD_BOUNDING_BOX = auto()



class ColorMode(StringEnum):
    """
    ColorMode: Color setting mode.

    DIRECT (default mode) allows each shape to be set arbitrarily

    CYCLE allows the color to be set via a color cycle over an attribute

    COLORMAP allows color to be set via a color map over an attribute
    """

    DIRECT = auto()
    CYCLE = auto()
    COLORMAP = auto()


class Box:
    """Box: Constants associated with the vertices of the interaction box"""

    WITH_HANDLE = [0, 1, 2, 3, 4, 5, 6, 7, 9]
    WITHOUT_HANDLE = list(range(8))
    LINE_HANDLE = [7, 6, 4, 2, 0, 7]
    LINE = [0, 2, 4, 6, 0]
    TOP_LEFT = 0
    TOP_CENTER = 7
    LEFT_CENTER = 1
    BOTTOM_RIGHT = 4
    BOTTOM_LEFT = 2
    CENTER = 8
    HANDLE = 9
    LEN = 8


BACKSPACE = 'delete' if sys.platform == 'darwin' else 'backspace'



import cv2
import numpy as np

from qtpy.QtCore import QEvent, Qt, QObject
from qtpy.QtGui import QImage, QPixmap
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QHBoxLayout, QLabel, QListWidgetItem, QListWidget, QMenu, \
    QComboBox, QPushButton, QStackedWidget, QDockWidget, QFileDialog, QSpinBox
from napari import Viewer
from napari.layers import Image, Labels
from napari.layers.labels._labels_constants import Mode
from scipy.ndimage import find_objects

from .projections import SliceDisplayWidget
from ..boundingbox import BoundingBoxLayer

import itertools
import csv

import PIL.Image as PilImage


def slice_len(slice_, count=None):
    if count is None:
        count = slice_.stop
    return len(range(*slice_.indices(count)))


class QObjectWidget(QWidget):
    def __init__(self, icon=None, name="Object", index=0, bounding_box=None, parent=None):
        super().__init__()
        self.textQVBoxLayout = QVBoxLayout()
        self.textUpQLabel = QLineEdit()
        self.textUpQLabel.setToolTip("Object name")
        self.id_label = QLabel()
        self.id_label.setText("Id: " + str(index))
        self.id_label.setToolTip("label index")
        self.textDownQLabel = QLabel()
        self.textQVBoxLayout.addWidget(self.textUpQLabel)
        self.textQVBoxLayout.addWidget(self.id_label)
        self.allQHBoxLayout  = QHBoxLayout()
        self.iconQLabel      = QLabel()
        self.allQHBoxLayout.addWidget(self.iconQLabel, 0)
        self.allQHBoxLayout.addLayout(self.textQVBoxLayout, 1)
        self.setLayout(self.allQHBoxLayout)
        if icon is not None:
            self.iconQLabel.setPixmap(icon)
        self.setTextUp(name)
        if bounding_box is not None:
            self.setTextDown(bounding_box)
        self.name = name
        self.parent = parent

    @property
    def name(self):
        return self.textUpQLabel.text()

    @name.setter
    def name(self, name):
        self.textUpQLabel.setText(name)

    def setTextUp(self, text):
        self.textUpQLabel.setText(text)

    def setTextDown(self, bb):
        self.textDownQLabel.setText(("corner 1: %s\ncorner 2: %s" % tuple(bb)))

    def setIcon(self, icon):
        self.iconQLabel.setPixmap(icon)

    def bounding_box_str(self, bb):
        pass


class QObjectListWidgetItem(QListWidgetItem):
    def __init__(self, name, bounding_box, index, parent, viewer, image_layer, mask_layer, channels_dim, *args):
        super().__init__(None, *args)
        if name is None:
            name = parent.name_template
        self.object_item = QObjectWidget(name=name, index=index, bounding_box=bounding_box, parent=self)
        self.setSizeHint(self.object_item.sizeHint())
        self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)
        self.parent = parent
        self.viewer = viewer
        self.image_layer = image_layer
        self.mask_layer = mask_layer
        self.bounding_box = bounding_box
        self.channels_dim = channels_dim

    def __hash__(self):
        return id(self)

    @property
    def idx(self):
        return int(self.object_item.id_label.text().split(":")[-1])

    @idx.setter
    def idx(self, idx):
        self.object_item.id_label.setText("Id: " + str(idx))

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, icon):
        self._icon = icon.reshape(icon.shape[0], icon.shape[1], -1)
        if self._icon.shape[-1] == 1:
            self._icon = np.tile(self._icon, (1, 1, 3))
        elif self._icon.shape[-1] == 4:
            self._icon = self._icon[..., :-1]
        scale = 64 / max(self._icon.shape[1], self._icon.shape[0])
        out_size = (int(self._icon.shape[1]*scale), int(self._icon.shape[0]*scale))
        self._resized_icon = cv2.resize(self._icon, out_size, interpolation=cv2.INTER_LANCZOS4)
        img = QImage(self._resized_icon, self._resized_icon.shape[1], self._resized_icon.shape[0], self._resized_icon.shape[1] * 3, QImage.Format.Format_RGB888)
        self.pixmap = QPixmap()
        self.pixmap = self.pixmap.fromImage(img, Qt.ImageConversionFlag.ColorOnly)
        self.object_item.setIcon(self.pixmap)
        # self.update_icon()

    @property
    def bounding_box(self):
        return self._bounding_box

    @bounding_box.setter
    def bounding_box(self, bounding_box):
        self._bounding_box = bounding_box
        self.object_item.setTextDown(bounding_box)
        self.object_item.update()
        if self.parent.crop_image_layer is not None and \
            self.parent.selected_idx == self.parent.indexFromItem(self).row():
            self.parent.crop_image_layer.data = self.image_layer.data[self.bbox_idx]
            self.parent.crop_image_layer.translate = tuple(self.bounding_box[0, :])
            self.parent.crop_image_layer.refresh()
            self.parent.crop_mask_layer.data = self.mask_layer.data[self.bbox_idx]
            self.parent.crop_mask_layer.translate = tuple(self.bounding_box[0, :])
            self.parent.crop_mask_layer.refresh()
        self.update_icon()

    @property
    def bbox_idx(self):
        bbox_idx = tuple(
            slice(self.bounding_box[0, d], self.bounding_box[1, d]) for d in range(self.bounding_box.shape[1]))
        return bbox_idx

    @property
    def name(self):
        return self.object_item.name

    @name.setter
    def name(self, name):
        self.object_item.name = name

    def create_layers(self, colormap):
        image_layer = Image(self.image_layer.data[self.bbox_idx], name="Cropped Image %d" % self.idx, colormap=colormap,
                            rgb=self.image_layer.rgb)
        image_layer.translate = self.bounding_box[0, :]+self.image_layer.translate
        self.viewer.add_layer(image_layer)
        mask_layer = Labels(self.mask_layer.data[self.bbox_idx], name="Cropped Mask %d" % self.idx)
        mask_layer.selected_label = self.idx
        mask_layer.translate = self.bounding_box[0, :]+self.mask_layer.translate
        mask_layer.brush_size = 1
        mask_layer.mode = Mode.PAINT
        self.viewer.add_layer(mask_layer)
        return image_layer, mask_layer

    def unselect(self):
        if self.parent.crop_image_layer is not None:
            if self.parent.crop_image_layer in self.viewer.layers:
                self.viewer.layers.remove(self.parent.crop_image_layer)
            self.parent.crop_image_layer = None
        if self.parent.crop_mask_layer is not None:
            if self.parent.crop_mask_layer in self.viewer.layers:
                self.viewer.layers.remove(self.parent.crop_mask_layer)
            self.parent.crop_mask_layer.mouse_drag_callbacks.remove(QObjectListWidgetItem.update_layer)
            self.parent.crop_mask_layer = None
        if self.parent.projections_widget is not None:
            self.viewer.window.remove_dock_widget(self.parent.projections_widget)
            self.parent.projections_widget = None
        self.mask_layer.visible = True
        self.image_layer.visible = True

    def on_select(self):
        self.unselect()
        self.parent.crop_image_layer, self.parent.crop_mask_layer = self.create_layers(self.image_layer.colormap)
        self.parent.crop_image_layer.name = self.parent.crop_image_layer.name
        self.parent.crop_mask_layer.name = self.parent.crop_mask_layer.name
        self.parent.crop_mask_layer.events.set_data.connect(self.on_data_change)
        if self.image_layer.ndim > 2:
            self.parent.projections_widget = SliceDisplayWidget(self.viewer, self.parent.crop_image_layer, self.parent.crop_mask_layer, self.channels_dim)
            qt_widget = self.viewer.window.add_dock_widget(self.parent.projections_widget)
            qt_widget.setFloating(True)
            qt_widget.setFeatures(qt_widget.features() & ~QDockWidget.DockWidgetClosable)
            def set_dockable(state):
                qt_widget.setAllowedAreas(Qt.AllDockWidgetAreas if state else Qt.NoDockWidgetArea)
            self.parent.projections_widget.dockable_checkbox.clicked.connect(set_dockable)
        self.parent.crop_mask_layer.mouse_drag_callbacks.append(self.update_layer)
        self.viewer.layers.selection.select_only(self.parent.crop_mask_layer)
        self.mask_layer.visible = False
        self.image_layer.visible = False

    def on_data_change(self, event):
        if self.channels_dim is not None and not self.image_layer.rgb:
            if self.parent.crop_mask_layer._mode == Mode.PAINT:
                data = self.parent.crop_mask_layer.data.max(axis=self.channels_dim, keepdims=True)
                self.parent.crop_mask_layer.data[:] = data
            elif self.parent.crop_mask_layer._mode == Mode.ERASE:
                data = self.parent.crop_mask_layer.data.min(axis=self.channels_dim, keepdims=True)
                self.parent.crop_mask_layer.data[:] = data
        self.mask_layer.data[self.bbox_idx] = self.parent.crop_mask_layer.data
        if self.parent.projections_widget is not None:
            self.parent.projections_widget.update_slider_ranges()
        self.parent.on_layer_event(event)
    @staticmethod
    def update_layer(layer, event):
        yield
        while event.type == "mouse_move":
            yield
        layer.refresh()

    def update_icon(self):
        visible_dims = list(self.viewer.dims.displayed)[-2:]
        bbox_idx = tuple(
            slice(max(self.bounding_box[0, d], 0), self.bounding_box[1, d]) if d in visible_dims
            else self.bounding_box[:, d].mean().astype(int)
            for d in range(self.bounding_box.shape[1])
        )
        if self.image_layer is None:
            icon = np.zeros([len(bbox_idx[d]) for d in range(len(bbox_idx)) if type(bbox_idx[d]) is slice])
        elif self.image_layer.dtype == np.uint16:
            icon = (self.image_layer.data[bbox_idx] // (2 ** 8)).astype(np.uint8)
        else:
            icon = self.image_layer.data[bbox_idx].astype(np.uint8)
        if self.image_layer is not None and not self.image_layer.rgb:
            icon = (self.image_layer.colormap.colors[icon] * 255).astype(np.uint8)[..., :3]
        else:
            icon = icon.astype(np.uint8)[..., :3]
        if self.mask_layer is not None:
            overlay = self.mask_layer.data[bbox_idx] == self.idx
            img = PilImage.fromarray(icon).convert(mode="RGBA")
            mask = PilImage.fromarray(overlay).convert(mode="RGBA")
            mask.putalpha(PilImage.fromarray((overlay*127).astype(np.uint8)))
            img.alpha_composite(mask)
            icon = np.asarray(img)
        self.icon = icon


class ObjectListWidget(QListWidget):
    def __init__(self, viewer, name_template, bounding_box_layer, image_layer, mask_layer, channels_dim, indices=None):
        super().__init__()
        self._bounding_box_layer = None
        self._image_layer = None
        self._mask_layer = None
        self.crop_image_layer = None
        self.crop_mask_layer = None
        self.projections_widget = None
        self._mouse_down = False
        self.selected_idx = None
        self.previously_hovered = None
        self.previous_face_color = None
        self.previous_edge_color = None
        self.bb_was_visible = None
        if indices is None:
            indices = range(1, 1000000)
        self.indices = itertools.cycle(indices)
        self.current_index = 0
        self.object_counter = 0
        self.viewer = viewer
        self.name_template = name_template
        self.channels_dim = channels_dim
        self.image_layer = image_layer
        self.mask_layer = mask_layer

        self.bounding_box_layer = bounding_box_layer
        self.setObjectName("Objects")
        self.update_items()
        self.itemClicked.connect(self.select_item)
        self.installEventFilter(self)
        self.viewer.bind_key('d')(self.toggle_bb_visibility)
        # bounding_box_layer.events.set_data.connect(on_data_change)

    def __del__(self):
        self._bounding_box_layer.mouse_drag_callbacks.remove(self.bounding_box_change)

    @property
    def bounding_box_layer(self):
        return self._bounding_box_layer

    @bounding_box_layer.setter
    def bounding_box_layer(self, new_layer):
        if self._bounding_box_layer:
            self._bounding_box_layer.mouse_drag_callbacks.remove(self.bounding_box_change)
            self._bounding_box_layer.mouse_double_click_callbacks.remove(self._on_bb_double_click)
            self._bounding_box_layer.events.disconnect(self.on_layer_event)
        self._bounding_box_layer = new_layer
        self._bounding_box_layer.mouse_drag_callbacks.append(self.bounding_box_change)
        self._bounding_box_layer.mouse_double_click_callbacks.append(self._on_bb_double_click)
        self._bounding_box_layer.events.connect(self.on_layer_event)
        self._bounding_box_layer.current_properties |= {"label": self.current_index}
        self._bounding_box_layer.text = {
            "text": "{label:d}",
            "size": 10,
            "color": "green"
        }
        self.update_items()

    @property
    def image_layer(self):
        return self._image_layer

    @image_layer.setter
    def image_layer(self, new_layer):
        self._image_layer = new_layer
        for i in range(self.count()):
            item = self.item(i)
            item.image_layer = new_layer
            item.update_icon()
        self.update_items()

    @property
    def mask_layer(self):
        return self._mask_layer

    @mask_layer.setter
    def mask_layer(self, new_layer):
        self._mask_layer = new_layer
        for i in range(self.count()):
            item = self.item(i)
            item.mask_layer = new_layer
            item.update_icon()
            if self.currentIndex().row() == i:
                item.on_select()
        self.update_items()

    def on_layer_event(self, event):
        # print(event.type, self._mouse_down)
        if event.type == "set_data" and not self._mouse_down:
            self.update_items(True)

    def select_item(self, item):
        item.on_select()
        selected_idx = self.indexFromItem(item).row()
        if selected_idx == self.selected_idx:
            self.selected_idx = None
            item.unselect()
            self.clearSelection()
            self.clearFocus()
            self.bounding_box_layer.edge_color[selected_idx] = self.previous_edge_color
            self.bounding_box_layer.data = self.bounding_box_layer.data
        else:
            if self.selected_idx is not None:
                self.bounding_box_layer.edge_color[self.selected_idx] = self.previous_edge_color
            self.selected_idx = selected_idx
            self.previous_edge_color = self.bounding_box_layer.edge_color[selected_idx].copy()
            self.bounding_box_layer.edge_color[selected_idx] = (1., 0., 0., 1.)
            self.bounding_box_layer.data = self.bounding_box_layer.data

    def bounding_box_corners(self):
        if self._bounding_box_layer is None or len(self._bounding_box_layer.data) == 0:
            return []
        b_data = np.round(np.asarray(self._bounding_box_layer.data)).astype(int)
        if self.image_layer is not None:
            b_data -= self.image_layer.translate.astype(int)
        return np.concatenate([b_data.min(1, keepdims=True), b_data.max(1, keepdims=True)], axis=1)

    def eventFilter(self, source: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Close:
            if self.crop_image_layer is not None:
                self.viewer.layers.remove(self.crop_image_layer)
                self.viewer.layers.remove(self.crop_mask_layer)
                self.viewer.window.remove_dock_widget(self.projections_widget)
        elif event.type() == QEvent.ContextMenu and source is self:
            menu = QMenu()
            menu.addAction("Create layers")
            if menu.exec_(event.globalPos()):
                item = self.itemAt(event.pos())
                item.create_layers(self.viewer, self.image_layer.colormap)
            return True
        elif event.type() == QEvent.Enter and type(source) == QObjectWidget:
            item = source.parent
            if item is None:
                return False
            idx = self.indexFromItem(item).row()
            self.previously_hovered = idx
            self.previous_face_color = self.bounding_box_layer.face_color[idx].copy()
            self.bounding_box_layer.face_color[idx] = (1., 1.0, 1.0, 0.5)
            self.bounding_box_layer.data = self.bounding_box_layer.data
            return True
        elif event.type() == QEvent.Leave and type(source) == QObjectWidget:
            if self.previous_face_color is not None:
                self.bounding_box_layer.face_color[self.previously_hovered] = self.previous_face_color
                self.bounding_box_layer.data = self.bounding_box_layer.data
                self.previously_hovered = None
                self.previous_face_color = None
            return True
        return super().eventFilter(source, event)

    def toggle_bb_visibility(self, _=None):
        if self.bounding_box_layer is not None:
            self.bounding_box_layer.visible = not self.bounding_box_layer.visible

    def bounding_box_change(self, layer, event):
        previous_data = np.asarray([bb.copy() for bb in self.bounding_box_layer.data])
        self._mouse_down = True
        print(len(previous_data), len(layer.data))
        yield
        if len(layer.data)>len(previous_data):
            self.bounding_box_layer.features["label"].iat[-1] = next(self.index())
            text = dict(self.bounding_box_layer.text)
            del text["values"]
            text["text"] = "{label:d}"
            self.bounding_box_layer.text = text
        while event.type == "mouse_move":
            yield
        self._mouse_down = False
        new_data = np.asarray(self.bounding_box_layer.data)
        if len(new_data) > 0 and (np.any(np.all(new_data > self.image_layer.data.shape, 1))\
                or np.any(np.all(new_data < 0, 1))):
            self.bounding_box_layer.data = previous_data
        if np.shape(previous_data) != np.shape(self.bounding_box_layer.data):
            if len(new_data) > len(previous_data):
                self.bounding_box_layer.data = [np.clip(data, 0, np.asarray(self.image_layer.data.shape) - 1) for data in new_data]
            else:
                removed_idx = np.argwhere(~np.any(np.equal(previous_data[np.newaxis], new_data[:, np.newaxis]), (-2, -1)))
                print("removed_idx was ", removed_idx)
                self.takeItem(removed_idx)
            '''if self.crop_mask_layer is not None:
                self.viewer.layers.remove(self.crop_mask_layer)
                self.viewer.layers.remove(self.crop_image_layer)
                self.crop_mask_layer = None
                self.crop_image_layer = None
                self.viewer.window.remove_dock_widget(self.projections_widget)'''
            self.update_items()
            if self.projections_widget is not None:
                for p in self.projections_widget.projections:
                    p.update()
        elif not np.allclose(previous_data, self.bounding_box_layer.data):
            idx = int(np.argwhere(~np.all(np.isclose(previous_data, self.bounding_box_layer.data), (1, 2))))
            item = self.item(idx)
            if item is None:
                return
            self.item(idx).bounding_box = self.bounding_box_corners()[idx]
            self.item(idx).update_icon()
            if self.projections_widget is not None:
                for p in self.projections_widget.projections:
                    p.update_overlay()
                    p.update()

    def update_items(self, update_all_icons=False):
        bounding_boxes = self.bounding_box_corners()
        if len(bounding_boxes) == 0:
            self.clear()
        elif len(bounding_boxes) < self.count():
            previous_data = np.asarray(list(np.unique(self.item(i).bounding_box, axis=0) for i in range(self.count())))
            removed_idx = np.argwhere(np.all(np.any(~np.equal(previous_data[np.newaxis], bounding_boxes[:, np.newaxis]), (-2, -1)), axis=0))
            self.takeItem(np.squeeze(removed_idx))
        for i, bb in enumerate(bounding_boxes):
            curr_item = self.item(i)
            if curr_item is not None:
                if not np.allclose(curr_item.bounding_box, bb):
                    curr_item.bounding_box = bb
                elif curr_item.image_layer is not self.image_layer:
                    curr_item.image_layer = self.image_layer
                elif update_all_icons:
                    curr_item.update_icon()
            else:
                new_item = QObjectListWidgetItem(self.name_template, bb, self.bounding_box_layer.features["label"][i], self, self.viewer, self.image_layer, self.mask_layer, self.channels_dim)
                self.insertItem(i, new_item)

    def insertItem(self, index, item: QObjectListWidgetItem):
        super().insertItem(index, item)
        self.setItemWidget(item, item.object_item)

    def addItem(self, item: QObjectListWidgetItem):
        self.insertItem(self.count(), item)

    def index(self):
        while True:
            for idx in self.indices:
                yield idx

    def _on_bb_double_click(self, layer=None, event=None):
        if len(self.bounding_box_layer.selected_data) == 0:
            return
        idx = next(iter(self.bounding_box_layer.selected_data))
        item = self.item(idx)
        self.setCurrentItem(item)
        self.select_item(item)


class ListWidgetBB(QWidget):
    def __init__(self, viewer: Viewer):
        super().__init__()
        layout = QVBoxLayout()
        self.viewer = viewer
        self.prev_n_layers = len(viewer.layers)
        self.channels_dim = None
        self.list_widget = None
        self.prev_bb_index = 0
        self.prev_img_index = 0
        self.prev_mask_index = 0
        viewer.layers.events.connect(self.update_layers)
        self.bounding_box_layer_dropdown = QComboBox()
        self.bounding_box_layer_dropdown.addItem("[Bounding box layer]")
        self.bounding_box_layer_dropdown.currentIndexChanged.connect(self.bb_index_change)
        self.image_layer_dropdown = QComboBox()
        self.image_layer_dropdown.addItem("[Image layer]")
        self.image_layer_dropdown.setCurrentIndex(0)
        self.image_layer_dropdown.currentIndexChanged.connect(self.img_index_change)
        self.mask_layer_dropdown = QComboBox()
        self.mask_layer_dropdown.addItem("[Label layer]")
        self.mask_layer_dropdown.setCurrentIndex(0)
        self.mask_layer_dropdown.currentIndexChanged.connect(self.mask_index_change)
        self.update_layers()
        layout.addWidget(self.bounding_box_layer_dropdown)
        layout.addWidget(self.image_layer_dropdown)
        layout.addWidget(self.mask_layer_dropdown)
        self.next_index_spinner = QSpinBox()
        self.next_index_spinner.setMinimum(1)
        self.next_index_spinner.setMaximum(1000000)
        self.next_index_spinner.setToolTip("Next index")
        self.next_index_spinner.setVisible(False)
        layout.addWidget(self.next_index_spinner)
        self.list_widget_container = QStackedWidget()
        no_data_widget = QWidget()
        no_data_layout = QVBoxLayout()
        no_data_layout.addWidget(QLabel("No bounding box layer selected"))
        add_bb_button = QPushButton("Create Bounding Box layer")
        add_bb_button.clicked.connect(self.create_bb_layer)
        no_data_layout.addWidget(add_bb_button)
        no_data_widget.setLayout(no_data_layout)
        self.list_widget_container.addWidget(no_data_widget)
        self.list_widget_container.setCurrentIndex(0)
        layout.addWidget(self.list_widget_container)
        buttons_widget = QWidget()
        buttons_layout = QHBoxLayout()
        import_button = QPushButton("Import")
        import_button.clicked.connect(self.import_bounding_boxes)
        export_button = QPushButton("Export")
        export_button.clicked.connect(self.export_bounding_boxes)
        from_labels_button = QPushButton("From labels")
        from_labels_button.clicked.connect(self.bounding_boxes_from_labels)
        buttons_layout.addWidget(import_button)
        buttons_layout.addWidget(export_button)
        buttons_layout.addWidget(from_labels_button)
        buttons_widget.setLayout(buttons_layout)
        layout.addWidget(buttons_widget)
        self.setLayout(layout)
        self.name_template = "Object"
        self.update_list()
        self.installEventFilter(self)

    def bb_index_change(self, index):
        if index == 0 and self.bounding_box_layer_dropdown.count() > 0:
            self.bounding_box_layer_dropdown.setCurrentIndex(self.prev_bb_index)
            return
        self.prev_bb_index = index
        self.update_list()

    def mask_index_change(self, index):
        if index == 0 and self.mask_layer_dropdown.count() > 0:
            self.mask_layer_dropdown.setCurrentIndex(self.prev_mask_index)
            return
        if self.list_widget is not None:
            self.list_widget.mask_layer = self.mask_layer
        self.prev_mask_index = index

    def img_index_change(self, index):
        if index == 0 and self.image_layer_dropdown.count() > 0:
            self.image_layer_dropdown.setCurrentIndex(self.prev_img_index)
            return
        self.prev_img_index = index
        self.update_channels_dim(index)

    def update_layers(self, event=None):
        type_ = event.type if event else "reordered"
        # print(type_)
        if type_ in ["reordered", "removed"]:
            bb_idx = 1
            img_idx = 1
            mask_idx = 1
            for layer in self.viewer.layers:
                if isinstance(layer, BoundingBoxLayer):
                    if bb_idx >= self.bounding_box_layer_dropdown.count():
                        self.bounding_box_layer_dropdown.addItem(layer.name)
                    else:
                        self.bounding_box_layer_dropdown.setItemText(bb_idx, layer.name)
                    if self.bounding_box_layer is None\
                            or self.bounding_box_layer_dropdown.itemText(bb_idx) == self.bounding_box_layer.name:
                        self.bounding_box_layer_dropdown.setCurrentIndex(bb_idx)
                    bb_idx += 1
                elif isinstance(layer, Image):
                    if img_idx >= self.image_layer_dropdown.count():
                        self.image_layer_dropdown.addItem(layer.name)
                    else:
                        self.image_layer_dropdown.setItemText(img_idx, layer.name)
                    if self.image_layer is None\
                            or self.image_layer_dropdown.itemText(img_idx) == self.image_layer.name:
                        self.image_layer_dropdown.setCurrentIndex(img_idx)
                    img_idx += 1
                elif isinstance(layer, Labels):
                    if mask_idx >=self.mask_layer_dropdown.count():
                        self.mask_layer_dropdown.addItem(layer.name)
                    else:
                        self.mask_layer_dropdown.setItemText(mask_idx, layer.name)
                    if self.mask_layer is None\
                            or self.mask_layer_dropdown.itemText(mask_idx) == self.mask_layer.name:
                        self.mask_layer_dropdown.setCurrentIndex(mask_idx)
                    mask_idx += 1
            if type_ == "removed":
                while bb_idx < self.bounding_box_layer_dropdown.count():
                    self.bounding_box_layer_dropdown.removeItem(bb_idx)
                while img_idx < self.image_layer_dropdown.count():
                    self.image_layer_dropdown.removeItem(img_idx)
                while mask_idx < self.mask_layer_dropdown.count():
                    self.mask_layer_dropdown.removeItem(mask_idx)
        elif type_ == "inserted":
            layer = self.viewer.layers[-1]
            if isinstance(layer, BoundingBoxLayer):
                self.bounding_box_layer_dropdown.addItem(layer.name)
                if self.bounding_box_layer_dropdown.count() == 2:
                    self.bounding_box_layer_dropdown.setCurrentIndex(1)
            elif isinstance(layer, Image):
                self.image_layer_dropdown.addItem(layer.name)
                if self.image_layer_dropdown.count() == 2:
                    self.image_layer_dropdown.setCurrentIndex(1)
            elif isinstance(layer, Labels):
                self.mask_layer_dropdown.addItem(layer.name)
                if self.mask_layer_dropdown.count() == 2:
                    self.mask_layer_dropdown.setCurrentIndex(1)
        elif type_ == "name":
            layer = event.source[event.index]
            if isinstance(layer, BoundingBoxLayer):
                self.bounding_box_layer_dropdown.setItemText(event.index+1, layer.name)
            elif isinstance(layer, Image):
                self.image_layer_dropdown.setItemText(event.index+1, layer.name)
            elif isinstance(layer, Labels):
                self.mask_layer_dropdown.setItemText(event.index+1, layer.name)


    @property
    def bounding_box_layer(self):
        return self.viewer.layers[self.bounding_box_layer_dropdown.currentText()] \
            if self.bounding_box_layer_dropdown.currentIndex() > 0 else None

    @property
    def image_layer(self):
        return self.viewer.layers[self.image_layer_dropdown.currentText()] \
            if self.image_layer_dropdown.currentIndex() > 0 else None

    @property
    def mask_layer(self):
        return self.viewer.layers[self.mask_layer_dropdown.currentText()] \
            if self.mask_layer_dropdown.currentIndex() > 0 else None

    def create_bb_layer(self, _):
        if self.image_layer is not None:
            self.viewer.add_layer(
                BoundingBoxLayer(ndim=self.image_layer.ndim, edge_color="green", face_color="transparent"))

    def update_channels_dim(self, idx):
        if idx == 0 or self.image_layer is None:
            return
        smallest_dim = np.argsort(self.image_layer.data.shape)[0]
        if self.image_layer.data.shape[smallest_dim] <= 3:
            self.channels_dim = smallest_dim
        else:
            self.channels_dim = None
        if self.list_widget is not None:
            self.list_widget.image_layer = self.image_layer
            self.list_widget.channels_dim = self.channels_dim

    def update_list(self):
        if self.bounding_box_layer is not None:
            self.list_widget = ObjectListWidget(self.viewer, self.name_template, self.bounding_box_layer, self.image_layer, self.mask_layer, self.channels_dim, self.index_iterator())
            self.list_widget_container.addWidget(self.list_widget)
            self.list_widget_container.setCurrentIndex(1)
        else:
            self.list_widget_container.setCurrentIndex(0)

    def export_bounding_boxes(self):
        if self.bounding_box_layer is None:
            return
        filename = QFileDialog.getSaveFileName(self, "Select...", None, "(*.txt)")[0]
        bboxes = np.asarray(list(map(lambda bb: np.concatenate([bb.min(axis=0), bb.max(axis=0)]), self.bounding_box_layer.data)))
        with open(filename, "w") as f:
            writer = csv.writer(f)
            for i, bb in enumerate(bboxes):
                item = self.list_widget.item(i)
                writer.writerow([item.name, item.idx] + list(bb))

    def import_bounding_boxes(self):
        filename = QFileDialog.getOpenFileName(self, "Select...", None, "(*.txt)")[0]
        names = []
        idxs = []
        data = []
        with open(filename, "r") as f:
            reader = csv.reader(f)
            for line in reader:
                if len(line) == 0:
                    continue
                names.append(line[0])
                idxs.append(int(line[1]))
                data.append(list(map(lambda x: float(x), line[2:])))
        data = np.asarray(data)
        bounding_box_corners = np.reshape(data, (len(data), 2, -1))
        mask = np.asarray(list(itertools.product((False, True), repeat=bounding_box_corners[0].shape[1])))
        bounding_boxes = np.asarray([np.where(mask, bbc[1], bbc[0]) for bbc in bounding_box_corners])
        if self.bounding_box_layer is None:
            bounding_box_layer = BoundingBoxLayer(bounding_boxes, edge_color="green", face_color="transparent")
            self.viewer.add_layer(bounding_box_layer)
        else:
            self.bounding_box_layer.data = bounding_boxes

        for i, (name, idx) in enumerate(zip(names, idxs)):
            item = self.list_widget.item(i)
            item.name = name
            item.idx = idx
        self.next_index_spinner.setValue(max(idxs)+1)

    def bounding_boxes_from_labels(self):
        if self.mask_layer is None:
            return
        if self.bounding_box_layer is None:
            bb_layer = BoundingBoxLayer(ndim=self.image_layer.ndim, edge_color="green", face_color="transparent")
            self.viewer.add_layer(bb_layer)
        else:
            bb_layer = self.bounding_box_layer
        bb_corners = find_objects(self.mask_layer.data)
        ids = []
        bbs = []
        for i, bb in enumerate(bb_corners):
            if bb is None:
                continue
            min_ = [slice_.start for slice_ in bb]
            max_ = [slice_.stop - 1 for slice_ in bb]
            bb = np.asarray(np.where(list(itertools.product((False, True), repeat=3)), max_, min_))
            bbs.append(bb)
            ids.append(i+1)
        bb_layer.data = bbs
        for i, id_ in enumerate(ids):
            self.list_widget.item(i).idx = id_

    def index_iterator(self):
        while True:
            idx = self.next_index_spinner.value()
            self.next_index_spinner.setValue(idx+1)
            yield idx

    def eventFilter(self, a0: 'QObject', a1: 'QEvent') -> bool:
        if a0 == self and a1.type() == QEvent.Show:
            self.parent().setAllowedAreas(Qt.LeftDockWidgetArea)
            return True
        return super().eventFilter(a0, a1)



# A copy of napari.layers.shapes._shapes_key_bindings
from napari.layers.utils.layer_utils import register_layer_action
from napari.utils.translations import trans

from ._bounding_box_constants import Mode

def register_bounding_boxes_action(layer_type, description):
    return register_layer_action(layer_type, description)
'''
def register_bounding_boxes_actions(layer_type):

    @register_bounding_boxes_action(layer_type, trans._('Select bounding boxes'))
    def activate_bb_select_mode(layer):
        """Activate bounding box selection tool."""
        layer.mode = Mode.SELECT

    @register_bounding_boxes_action(layer_type, trans._('Pan/Zoom'))
    def activate_bb_pan_zoom_mode(layer):
        """Activate pan and zoom mode."""
        layer.mode = Mode.PAN_ZOOM


    @register_bounding_boxes_action(layer_type, trans._('Add bounding box'))
    def activate_add_bb_mode(layer):
        """Activate add bounding box tool."""
        layer.mode = Mode.ADD_BOUNDING_BOX
'''



import numpy as np
from cv2 import cv2
from napari import Viewer
from napari.layers import Labels
from scipy.interpolate import interp1d
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLabel, QSpinBox, QPushButton

def contour_cv2_mask_uniform(mask, contoursize_max):
    mask = mask.astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    areas = [cv2.contourArea(cnt) for cnt in contours]
    max_ind = np.argmax(areas)
    contour = np.squeeze(contours[max_ind])
    contour = np.reshape(contour, (-1, 2))
    contour = np.append(contour, contour[0].reshape((-1, 2)), axis=0)
    contour = contour.astype('float32')

    rows, cols = mask.shape
    delta = np.diff(contour, axis=0)
    s = [0]
    for d in delta:
        dl = s[-1] + np.linalg.norm(d)
        s.append(dl)

    if (s[-1] == 0):
        s[-1] = 1

    s = np.array(s) / s[-1]
    fx = interp1d(s, contour[:, 0] / rows, kind='linear')
    fy = interp1d(s, contour[:, 1] / cols, kind='linear')
    S = np.linspace(0, 1, contoursize_max, endpoint=False)
    X = rows * fx(S)
    Y = cols * fy(S)

    contour = np.transpose(np.stack([X, Y])).astype(np.float32)

    contour = np.stack((contour[:, 1], contour[:, 0]), axis=-1)
    return contour


class InterpolationWidget(QWidget):
    def __init__(self, viewer: Viewer):
        super().__init__()
        self.viewer = viewer
        layout = QVBoxLayout()
        self.active_labels_layer = None

        layout.addWidget(QLabel("dimension"))
        self.dimension_dropdown = QSpinBox()
        self.dimension_dropdown.setMinimum(0)
        self.dimension_dropdown.setMaximum(0)
        layout.addWidget(self.dimension_dropdown)

        layout.addWidget(QLabel("# contour points"))
        self.n_points = QSpinBox()
        self.n_points.setMinimum(10)
        self.n_points.setMaximum(1000)
        self.n_points.setValue(300)
        layout.addWidget(self.n_points)

        self.interpolate_button = QPushButton("Interpolate")
        self.interpolate_button.clicked.connect(self.interpolate)
        layout.addWidget(self.interpolate_button)
        self.viewer.layers.selection.events.connect(self.on_layer_selection_change)
        self.viewer.layers.events.connect(self.on_layers_event)
        viewer.dims.events.order.connect(self.on_order_change)
        viewer.dims.events.ndisplay.connect(lambda _: self.interpolate_button.setEnabled(viewer.dims.ndisplay == 2))
        self.setLayout(layout)
        layout.addStretch()
        self.on_layer_selection_change()

    def on_layer_selection_change(self, event=None):
        if event is None or event.type == "changed":
            active_layer = self.viewer.layers.selection.active
            print(isinstance(active_layer, Labels))
            if isinstance(active_layer, Labels):
                self.active_labels_layer = active_layer
                print(self.active_labels_layer.name)
                self.dimension_dropdown.setMaximum(self.active_labels_layer.ndim)

    def on_layers_event(self, event):
        if event.type == "inserted":
            print(event.type, event.source[event.index])

    def set_has_channels(self, has_channels):
        self.channels_dim_dropdown.setEnabled(has_channels)

    def on_order_change(self, event):
        extents = self.active_labels_layer.extent.data[1] + 1
        not_displayed = extents[list(self.viewer.dims.not_displayed)]
        ch_excluded = list(filter(lambda x: x > 3, not_displayed))
        if len(ch_excluded) == 0:
            new_dim = self.viewer.dims.not_displayed[0]
        else:
            new_dim = self.viewer.dims.order[int(np.argwhere(not_displayed == ch_excluded[0])[0])]
        self.dimension_dropdown.setValue(new_dim)

    def interpolate(self):
        print("interpolate")
        if self.active_labels_layer is None:
            return
        print("found layer")
        dimension = self.dimension_dropdown.value()
        n_contour_points = self.n_points.value()
        data = self.active_labels_layer.data
        layer_slice_template = [
            slice(None) if d in self.viewer.dims.displayed
                else None if d == dimension
                else self.viewer.dims.current_step[d]
            for d in range(self.active_labels_layer.ndim)]
        prev_cnt = None
        prev_layer = None
        for i in range(data.shape[dimension]):
            layer_slice = layer_slice_template.copy()
            layer_slice[dimension] = i
            mask = data[tuple(layer_slice)] == self.active_labels_layer.selected_label
            mask = mask.astype(np.uint8)
            if mask.max() == 0:
                continue
            cnt = contour_cv2_mask_uniform(mask, n_contour_points)
            centroid = cnt.mean(0)
            start_index = np.argmin(np.arctan2(*(cnt - centroid).T))
            cnt = np.roll(cnt, start_index, 0)
            if prev_cnt is not None:
                for j in range(prev_layer + 1, i):
                    inter_layer_slice = layer_slice_template.copy()
                    inter_layer_slice[dimension] = j
                    prev_w = i - j
                    cur_w = j - prev_layer
                    mean_cnt = (prev_w * prev_cnt + cur_w * cnt)/(prev_w + cur_w)
                    mean_cnt = mean_cnt.astype(np.int32)
                    mask = np.zeros_like(data[tuple(inter_layer_slice)])
                    cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, self.active_labels_layer.selected_label, -1)
                    data[tuple(inter_layer_slice)] = mask
            prev_cnt = cnt
            prev_layer = i
            self.active_labels_layer.refresh()


import itertools
import warnings

import numpy as np
from PIL import Image as PILImage
from PyQt5.Qt import QWIDGETSIZE_MAX
from PyQt5.QtCore import QSettings
from PyQt5.QtGui import QImage, QPixmap, QResizeEvent, QColor
from PyQt5.QtWidgets import QLabel, QSizePolicy, QWidget, QVBoxLayout, QSlider, QHBoxLayout, QGridLayout, QCheckBox, \
    QColorDialog, QPushButton
from napari.layers.labels._labels_constants import Mode
from qtpy import QtCore

from matplotlib import colors

class DataProjectionWidget(QLabel):
    def __init__(self, viewer, image_layer, mask_layer, displayed_axes, image_colormap=None, mask_colormap=None,
                 slices=None, flip_image=False, crosshair_color=None):
        super().__init__()
        self._crosshair_color = None
        self.viewer = viewer
        self.image_layer = image_layer
        self.mask_layer = mask_layer
        self.flip_image = flip_image
        self.max_width, self.max_height = QWIDGETSIZE_MAX, QWIDGETSIZE_MAX
        self.slices = list(slices) or [s//2 for s in self.image_data.shape]
        self.displayed_axes = displayed_axes
        overlay_shape = [self.image_data.shape[i] for i in range(self.image_data.ndim) if self.im_idx[i] == slice(None)][:2]
        self._overlay = np.zeros(overlay_shape, np.uint8)
        if image_layer.rgb and image_colormap is not None:
            warnings.warn("colormap provided for RGB image. Ignoring")
            self.image_colormap = None
        else:
            self.image_colormap = image_colormap
        self.mask_colormap = mask_colormap
        if crosshair_color is None:
            self.crosshair_color = (255, 0, 0, 160)
        else:
            self.crosshair_color = crosshair_color
        self.update()
        self.pixmap = None
        self.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.mousePressEvent = self.on_click
        self.setFrameStyle(QLabel.Raised | QLabel.Box)

    def setSlices(self, *args):
        assert len(args) % 2 == 0
        for i in range(0, len(args), 2):
            self.slices[args[i]] = max(min(args[i+1], self.image_data.shape[args[i]]-1), 0)
        self.update()

    @property
    def crosshair_color(self):
        return self._crosshair_color

    @crosshair_color.setter
    def crosshair_color(self, new_color):
        if all(map(lambda v: 0.<=v<=1., new_color)):
            new_color = map(lambda v: int(v*255), new_color)
        self._crosshair_color = tuple(new_color)
        self.update()

    @property
    def overlay(self):
        overlay = (np.ones(self._overlay.shape + (4,)) * self.crosshair_color).astype(np.uint8)
        overlay[..., -1] *= self._overlay
        if self.flip_image:
            overlay = np.transpose(overlay, (1, 0, 2))
        return PILImage.fromarray(overlay)

    @overlay.setter
    def overlay(self, new_overlay):
        self._overlay = new_overlay > 0

    def update_overlay(self, layer=None, coordinates=None):
        overlay_shape = [self.image_data.shape[i] for i in range(self.image_data.ndim) if
                         self.im_idx[i] == slice(None)][:2]
        overlay = np.zeros(overlay_shape, bool)
        if layer is not None and coordinates is not None:
            if layer._mode in [Mode.ERASE, Mode.PAINT]:
                if all(layer_dim == proj_dim for layer_dim, proj_dim in
                       zip(sorted(self.viewer.dims.displayed), sorted(self.displayed_axes))):
                    if 0 <= coordinates[self.displayed_axes[0]] < overlay.shape[1 if self.flip_image else 0]:
                        if self.flip_image:
                            overlay[:, coordinates[self.displayed_axes[0]]] = True
                        else:
                            overlay[coordinates[self.displayed_axes[0]]] = True
                    if 0 <= coordinates[self.displayed_axes[1]] < overlay.shape[0 if self.flip_image else 1]:
                        if self.flip_image:
                            overlay[coordinates[self.displayed_axes[1]]] = True
                        else:
                            overlay[:, coordinates[self.displayed_axes[1]]] = True
                else:
                    for dim in self.viewer.dims.displayed:
                        if dim == self.displayed_axes[0]:
                            start = max(coordinates[dim] - layer.brush_size // 2, 0)
                            end = min(max(coordinates[dim] + layer.brush_size // 2 + 1, 0), overlay.shape[1 if self.flip_image else 0])
                            if self.flip_image and coordinates[self.displayed_axes[1]] < overlay.shape[0]:
                                overlay[coordinates[self.displayed_axes[1]], start:end] = True
                            elif coordinates[self.displayed_axes[1]] < overlay.shape[1]:
                                overlay[start:end, coordinates[self.displayed_axes[1]]] = True
                        elif dim == self.displayed_axes[1]:
                            start = max(coordinates[dim] - layer.brush_size // 2, 0)
                            end = min(max(coordinates[dim] + layer.brush_size // 2 + 1, 0), overlay.shape[0 if self.flip_image else 1])
                            if self.flip_image and coordinates[self.displayed_axes[0]] < overlay.shape[1]:
                                overlay[start:end, coordinates[self.displayed_axes[0]]] = True
                            elif coordinates[self.displayed_axes[0]] < overlay.shape[0]:
                                overlay[coordinates[self.displayed_axes[0]], start:end] = True
        elif bool(layer) ^ bool(coordinates): # XOR
            raise ValueError("layer and coordinates should be both provided or left None")
        self.overlay = overlay

    @property
    def image_data(self):
        return self.image_layer.data

    @property
    def mask_data(self):
        return self.mask_layer.data

    @property
    def im_idx(self):
        return tuple(
                self.slices[i] if i not in self.displayed_axes
                    else slice(None) for i in range(self.image_layer.ndim))

    def update(self, update_icon=True, new_size=None):
        if update_icon:
            if any(self.im_idx[i] != slice(None) and self.im_idx[i] >= self.image_data.shape[i]
                    for i in range(self.image_data.ndim)):
                im_shape = tuple(self.image_data.shape[i] for i in range(self.image_data.ndim)
                                 if self.im_idx[i] == slice(None))[:2]
                im = np.zeros(im_shape + (4,), np.uint8)
                mask = np.zeros_like(im)
            else:
                print(self.im_idx)
                if self.image_colormap is None:
                    im = self.image_data[self.im_idx]
                else:
                    im = (self.image_colormap[self.image_data[self.im_idx].astype(np.uint16)//256]*255)

                mask = (self.mask_data[self.im_idx] > 0).astype(np.uint8)
                alpha = mask * 180
                mask = np.tile(mask[..., np.newaxis], (1, 1, 4))*255

                mask[..., -1] = alpha

                if self.flip_image:
                    im = np.transpose(im, (1, 0, 2))
                    mask = np.transpose(mask, (1, 0, 2))
            im = PILImage.fromarray(im.astype(np.uint8)).convert(mode="RGBA")
            mask = PILImage.fromarray(mask).convert(mode="RGBA")
            icon = PILImage.alpha_composite(im, mask)
            icon = PILImage.alpha_composite(icon, self.overlay).convert("RGB")
            icon = np.asarray(icon)
            _icon = icon.reshape(icon.shape[0], icon.shape[1], -1)
            if _icon.shape[-1] == 1:
                _icon = np.tile(_icon, (1, 1, 3))
            elif _icon.shape[-1] == 4:
                _icon = _icon[..., :-1]
            self.img = QImage(_icon, _icon.shape[1], _icon.shape[0],
                         _icon.shape[1] * 3, QImage.Format.Format_RGB888)
        pixmap = QPixmap()
        pixmap = pixmap.fromImage(self.img, QtCore.Qt.ImageConversionFlag.ColorOnly)
        self.setMinimumSize(64, 64)
        if new_size is not None:
            w = new_size.width()
            h = new_size.height()
        else:
            w = self.size().width()
            h = self.size().height()
        pixmap = pixmap.scaled(w, h, QtCore.Qt.KeepAspectRatio)
        self.setPixmap(pixmap)

    def on_resize(self, new_size):
        self.update(update_icon=False, new_size=new_size)

    def resizeEvent(self, a0: QResizeEvent) -> None:
        self.on_resize(a0.size())

    def on_click(self, _):
        order = tuple(filter(lambda x: x not in self.displayed_axes, self.viewer.dims.order))
        order = order + tuple(self.displayed_axes)
        self.viewer.dims.order = order
        self.viewer.dims.current_step = tuple(self.viewer.dims.current_step[dim] if dim in self.displayed_axes
                                         else self.slices[dim]+self.image_layer.translate[dim]
                                         for dim in range(len(self.viewer.dims.current_step)))


class SliceDisplayWidget(QWidget):
    def __init__(self, viewer, image_layer, mask_layer, channels_dim=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setObjectName("Projections")
        main_layout = QVBoxLayout()
        self.settings = QSettings()
        self.overlay_color = self.settings.value("projection_overlay_color", (255, 0, 0, 160))
        self.slider_labels = []
        self.sliders = []
        self.slider_widgets = []
        self.offset = image_layer.translate
        self.viewer = viewer
        self.image_layer = image_layer
        for dim in range(viewer.dims.ndim):
            label = QLabel("%d" % dim)
            slider = QSlider(QtCore.Qt.Horizontal)
            slider.setMinimum(0)
            slider.setMaximum(image_layer.data.shape[dim])
            slider.setSingleStep(1)
            slider.setFixedWidth(100)
            slider.valueChanged.connect(self.slider_callback(dim))
            slider_layout = QHBoxLayout()
            slider_layout.addWidget(label)
            slider_layout.addWidget(slider)
            slider_widget = QWidget()
            slider_widget.setLayout(slider_layout)
            main_layout.addWidget(slider_widget)
            self.slider_widgets.append(slider_widget)
            if dim not in viewer.dims.order[-viewer.dims.ndisplay:]:
                slider_widget.setVisible(False)
            self.sliders.append(slider)
        grid_layout = QGridLayout()
        self.projections = []
        for dim_pair in itertools.combinations(range(image_layer.ndim), 2):
            if channels_dim in dim_pair:
                continue
            if dim_pair[-2] == viewer.dims.order[-1] or dim_pair[-1] == viewer.dims.order[-2]:
                dim_pair = tuple(reversed(dim_pair))
                flip = True
            else:
                flip = False
            slices = [int(viewer.dims.current_step[dim]-image_layer.translate[dim]) if dim in image_layer._dims_not_displayed else 0 for dim in range(viewer.dims.ndim)]
            projection = DataProjectionWidget(viewer, image_layer, mask_layer, dim_pair, image_colormap=image_layer.colormap.colors, mask_colormap=mask_layer.colormap.colors, slices=slices, flip_image=flip, crosshair_color=self.overlay_color)
            grid_layout.setRowStretch(len(self.projections)//3, 1)
            grid_layout.setColumnStretch(len(self.projections)%3, 1)
            grid_layout.addWidget(projection, len(self.projections)//3, len(self.projections)%3, QtCore.Qt.AlignmentFlag.AlignCenter)
            self.projections.append(projection)
        self.grid_widget = QWidget()
        self.grid_widget.setLayout(grid_layout)
        self.grid_widget.setSizePolicy(self.grid_widget.sizePolicy().horizontalPolicy(), QSizePolicy.Expanding)
        main_layout.addWidget(self.grid_widget)
        extra_settings_layout = QHBoxLayout()
        self.dockable_checkbox = QCheckBox("dockable")
        self.dockable_checkbox.setChecked(True)
        extra_settings_layout.addWidget(self.dockable_checkbox)
        self.change_color_button = QPushButton()
        self.change_color_button.setFlat(True)
        self.change_color_button.setStyleSheet("background-color: %s;" % colors.to_hex(list(map(lambda v: v/255, self.overlay_color))))
        self.change_color_button.clicked.connect(self.show_overlay_colorpicker)
        self.change_color_button.setFixedWidth(20)
        self.change_color_button.setFixedHeight(20)
        self.setToolTip("Change overlay color")

        extra_settings_layout.addWidget(self.change_color_button)
        main_layout.addLayout(extra_settings_layout)
        self.setLayout(main_layout)
        self.setSizePolicy(self.grid_widget.sizePolicy().horizontalPolicy(), QSizePolicy.Expanding)
        viewer.dims.events.current_step.connect(self.on_step_change)
        viewer.dims.events.order.connect(self.on_order_change)
        mask_layer.mouse_drag_callbacks.append(self.update_layer)
        mask_layer.mouse_move_callbacks.append(self.show_crosshair)

    def show_crosshair(self, layer, event):
        if self.viewer.dims.ndisplay == 3:
            return

        float_coordinates = layer.world_to_data(event.position)
        coordinates = np.round(np.asarray(float_coordinates)).astype(int)
        for p in self.projections:
            p.update_overlay(layer, coordinates)
        for i, slider in enumerate(self.sliders):
            if i in self.viewer.dims.order[-self.viewer.dims.ndisplay:] and layer.data.shape[i] > coordinates[i] >= 0:
                slider.setValue(coordinates[i])

    def on_step_change(self, event=None, layers=None):
        for projection in self.projections:
            projection.setSlices(*itertools.chain(*((dim, int(event.source.current_step[dim]-self.offset[dim])) for dim in event.source.not_displayed)))

    def update_layer(self, layer, event):
        self.show_crosshair(layer, event)
        for projection in self.projections:
            projection.update()
        yield
        while event.type == "mouse_move":
            self.show_crosshair(layer, event)
            for projection in self.projections:
                projection.update()
            yield
        for projection in self.projections:
            projection.update()

    def update_slider_ranges(self):
        for dim in range(len(self.sliders)):
            self.sliders[dim].setMaximum(self.image_layer.data.shape[dim])

    def slider_callback(self, dim):
        def on_slider_change(val):
            for projection in self.projections:
                projection.setSlices(dim, val)
        return on_slider_change

    def on_order_change(self, _):
        for dim, slider in enumerate(self.slider_widgets):
            slider.setVisible(dim in self.viewer.dims.order[-self.viewer.dims.ndisplay:])
        for p in self.projections:
            if p.displayed_axes[-2] == self.viewer.dims.order[-1] or p.displayed_axes[-1] == self.viewer.dims.order[-2]:
                p.flip_image = not p.flip_image
                p.displayed_axes = type(p.displayed_axes)(reversed(p.displayed_axes))

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(640, 480)

    def show_overlay_colorpicker(self):
        color = QColorDialog.getColor(QColor(*self.overlay_color), options=QColorDialog.ShowAlphaChannel)
        if color.isValid():
            self.overlay_color = color.red(), color.green(), color.blue(), color.alpha()
            self.settings.setValue("projection_overlay_color", self.overlay_color)
            self.change_color_button.setStyleSheet("background-color: %s;" % colors.to_hex(list(map(lambda v: v/255, self.overlay_color))))
            for projection in self.projections:
                projection.crosshair_color = self.overlay_color